# # Example 4: Debugging
#
# Debugging is half the coding process in software, and in PyRTL, it's no different.
# PyRTL provides some additional challenges when it comes to debugging, as a problem may
# surface long after the error was made. Fortunately, PyRTL comes with various features
# to help you find mistakes.
import io
import random

import pyrtl
from pyrtl.rtllib.adders import kogge_stone
from pyrtl.rtllib.multipliers import tree_multiplier

# This example covers debugging strategies for PyRTL. For general Python debugging, we
# recommend healthy use of the `assert` statement, and use of `pdb` for tracking down
# bugs. However, PyRTL introduces some new complexities because the place where
# functionality is defined (when you construct and operate on PyRTL classes) is separate
# in time from where that functionality is executed (i.e. during `Simulation`). Thus,
# sometimes it hard to track down where a `WireVector` might have come from, or what
# exactly it is doing.
#
# In this example, we build a circuit that adds up three `Input` values. However,
# instead of building an add function ourselves or using the built-in "+" function in
# PyRTL, we will instead use the `kogge_stone()` adders in `rtllib`, PyRTL's standard
# library.
in1 = pyrtl.Input(8, "in1")
in2 = pyrtl.Input(8, "in2")
in3 = pyrtl.Input(8, "in3")

out = pyrtl.Output(10, "out")

add1_out = kogge_stone(in1, in2)
add2_out = kogge_stone(add1_out, in2)
out <<= add2_out

# The most basic way of debugging PyRTL is to connect a value to an output wire and use
# `Simulation` to trace the output. A simple `print` statement won't work because the
# `WireVectors` do not carry values while we are building the hardware.
#
# If we want to check the result of the first addition, we can connect an `Output` wire
# to the result wire of the first adder:
debug_out = pyrtl.Output(9, "debug_out")
debug_out <<= add1_out

# Now simulate the circuit.  Let's create some random inputs to feed our adder.
random.seed(93729473)  # used to make random calls deterministic for this example
vals1 = [random.randrange(256) for _ in range(20)]
vals2 = [random.randrange(256) for _ in range(20)]
vals3 = [random.randrange(256) for _ in range(20)]

sim = pyrtl.Simulation()
sim.step_multiple({"in1": vals1, "in2": vals2, "in3": vals3})

# In order to get the result data, you do not need to print a waveform of the trace. You
# always have the option to just pull the data out of the `Simulation`'s `tracer`
# directly:
print("---- Inputs and debug_out ----")
print("in1:       ", sim.tracer.trace["in1"])
print("in2:       ", sim.tracer.trace["in2"])
print("debug_out: ", sim.tracer.trace["debug_out"])
print("\n")

# Below, I am using the ability to directly retrieve the trace data to verify the
# correctness of the first adder
for cycle in range(len(vals1)):
    actual = sim.tracer.trace["debug_out"][cycle]
    expected = sim.tracer.trace["in1"][cycle] + sim.tracer.trace["in2"][cycle]
    assert actual == expected


# ## `probe()`
#
# Now that we have built some stuff, let's clear it so we can try again in a different
# way. We can start by clearing all of the hardware from the current working block with
# `reset_working_block()`. The working block is a global structure that keeps track of
# all the hardware you have built thus far. A "reset" will clear it so we can start
# fresh.
pyrtl.reset_working_block()

# In this example, we will be multiplying two numbers using `tree_multiplier()`. Again,
# create the two inputs and an output
in1 = pyrtl.Input(8, "in1")
in2 = pyrtl.Input(8, "in2")

out1 = pyrtl.Output(8, "out1")
out2 = pyrtl.Output(8, "out2")

multout = tree_multiplier(in1, in2)

# The following line will create a `probe()` named `std_probe` for later use, like an
# output.
pyrtl.probe(multout, "std_probe")

# We could also do the same thing during assignment. The next command will create a
# `probe()` (named `stdout_probe`) that refers to `multout`, and returns the wire
# `multout`. This achieves virtually the same thing as the `probe()` above, but it is
# done during assignment, so we skip a step by probing the wire before the
# multiplication. The `probe()` returns `multout`, the original wire, and `out1` will be
# assigned `multout * 2`.
out1 <<= pyrtl.probe(multout, "stdout_probe") * 2

# `probe()` can also be used with other operations like this:
pyrtl.probe(multout + 32, "adder_probe")

# or this:
pyrtl.probe(multout[2:7], "select_probe")

# or, similarly:
# (this will create a probe of `multout` while passing `multout[2:16]` to `out2`)
out2 <<= pyrtl.probe(multout)[2:16]  # notice probe names are not absolutely necessary

# As one can see, `probe()` can be used on any wire any time, such as before or during
# its operation, assignment, etc.
#
# Now on to the simulation... For variation, we'll recreate the random inputs:
vals1 = [random.randrange(10) for _ in range(10)]
vals2 = [random.randrange(10) for _ in range(10)]

sim = pyrtl.Simulation()
sim.step_multiple({"in1": vals1, "in2": vals2})

# Now we will show the values of the inputs and probes and look at that, we didn't need
# to make any outputs! (although `probe()` did).
print("---- Using Probes ----")
sim.tracer.render_trace(repr_func=str)
sim.tracer.print_trace()

# Say we wanted more information about one of those probes above at declaration. We
# could have used `set_debug_mode()` before their creation, like so:
print("\n--- Probe w/ debugging: ---")
pyrtl.set_debug_mode()
pyrtl.probe(multout - 16, "debugsubtr_probe")
pyrtl.set_debug_mode(debug=False)


# ## `WireVector` Stack Trace
#
# Another case that might arise is that a certain wire is causing an error to occur in
# your program. `WireVector` Stack Traces allow you to find out more about where a
# particular `WireVector` was made in your code. With this enabled the `WireVector` will
# store exactly were it was created, which should help with issues where there is a
# problem with an identified wire.
#
# Like above, just add the following line before the relevant `WireVector` might be made
# or at the beginning of the program.
pyrtl.set_debug_mode()

test_out = pyrtl.Output(9, "test_out")
test_out <<= kogge_stone(in1, in2)

# Now to retrieve information
wire_trace = test_out.init_call_stack

# This data is generated using the `traceback.format_stack()` call from the Python
# standard library's `traceback` module (look at the Python standard library docs for
# details on the function). Therefore, the stack traces are stored as a list with the
# outermost call first.
print("---- Stack Trace ----")
for frame in wire_trace:
    print(frame)

# ### Storage of Additional Debug Data
#
# WARNING: the debug information generated by the following two processes are not
# guaranteed to be preserved when functions (eg. `synthesize()`) are done over the
# `Block`.
#
# However, if the stack trace does not give you enough information about the
# `WireVector`, you can also embed additional information into the wire itself. Two ways
# of doing so are by changing the name of the `WireVector`, or by adding your own custom
# metadata to the `WireVector`.
#
# So far, each `Input` and `Output` WireVector` have been given their own names, but
# normal `WireVectors` can also be given names by supplying the name argument to the
# constructor.
dummy_wv = pyrtl.WireVector(1, name="blah")

# A `WireVector`'s `name` can also be changed later:
dummy_wv.name = "argh"

# Also, because of the flexible nature of Python, you can also add custom properties to
# the `WireVector`:
dummy_wv.my_custom_property_name = "John Clow is great"
dummy_wv.custom_value_028493 = 13

# Remove the `WireVector` from the `Block` to prevent problems with the rest of this
# example.
pyrtl.working_block().remove_wirevector(dummy_wv)

# ## Trivial Graph Format
#
# Finally, there is a handy way to view your hardware creations as a graph. The function
# `output_to_trivialgraph()` will render your hardware in a format that you can then
# open with the free software "yEd" (http://en.wikipedia.org/wiki/YEd). There are
# options under the "hierarchical" rendering to draw something that looks quite like a
# circuit.
#
# Also see `output_to_svg()`.
pyrtl.working_block().sanity_check()

# So that `output_to_trivial_graph()` will work.
pyrtl.passes._remove_unused_wires(pyrtl.working_block())

print("\n--- Trivial Graph Format (first 10 lines) ---")
with io.StringIO() as tgf:
    pyrtl.output_to_trivialgraph(tgf)
    for i, line in enumerate(tgf.getvalue().split("\n")):
        if i == 10:
            break
        print(line)

    print("...")



# # Example 5: Making use of PyRTL and Introspection.
import pyrtl


# The following example shows how PyRTL can be used to make some interesting hardware
# structures using Python introspection. In particular, this example makes a N-stage
# pipeline structure. Any specific pipeline is then a derived class of `SimplePipeline`
# where methods with names starting with `stage` are stages, and new members with names
# not starting with `_` are to be registered for the next stage.
#
# ## Pipeline builder with auto generation of pipeline registers.
class SimplePipeline:
    def __init__(self):
        self._pipeline_register_map = {}
        self._current_stage_num = 0
        stage_list = [method for method in dir(self) if method.startswith("stage")]
        for stage in sorted(stage_list):
            stage_method = getattr(self, stage)
            stage_method()
            self._current_stage_num += 1

    def __getattr__(self, name):
        try:
            return self._pipeline_register_map[self._current_stage_num][name]
        except KeyError as exc:
            msg = (
                f'error, no pipeline register "{name}" defined for stage '
                f"{self._current_stage_num}"
            )
            raise pyrtl.PyrtlError(msg) from exc

    def __setattr__(self, name, value):
        if name.startswith("_"):
            # do not do anything tricky with variables starting with '_'
            object.__setattr__(self, name, value)
        else:
            next_stage = self._current_stage_num + 1
            pipereg_id = f"{self._current_stage_num} to {next_stage}"
            rname = f"pipereg_{pipereg_id}_name"
            new_pipereg = pyrtl.Register(bitwidth=len(value), name=rname)
            if next_stage not in self._pipeline_register_map:
                self._pipeline_register_map[next_stage] = {}
            self._pipeline_register_map[next_stage][name] = new_pipereg
            new_pipereg.next <<= value


# ## A very simple pipeline to show how registers are inferred.
class SimplePipelineExample(SimplePipeline):
    def __init__(self):
        self._loopback = pyrtl.WireVector(1, "loopback")
        super().__init__()

    def stage0(self):
        self.n = ~self._loopback

    def stage1(self):
        self.n = self.n

    def stage2(self):
        self.n = self.n

    def stage3(self):
        self.n = self.n

    def stage4(self):
        self._loopback <<= self.n


simplepipeline = SimplePipelineExample()
print(pyrtl.working_block())
# ## Simulation of the core
sim = pyrtl.Simulation()
sim.step_multiple({}, nsteps=15)
sim.tracer.render_trace()



"""Convert a PyRTL example script to a Jupyter notebook.

Usage::

    python to_ipynb.py example1-combologic.py example1-combologic.ipynb

This converts comment blocks to Markdown cells, and code blocks to code cells, with some
PyRTL-specific transformations:

- Add `%pip install` magic to install PyRTL in the notebook environment.

- Add a `reset_working_block()` call to reset PyRTL's state when re-running the
  notebook.

- Convert some code references to documentation hyperlinks. For example, `WireVector`
  will be replaced with
  [WireVector](https://pyrtl.readthedocs.io/en/latest/basic.html#pyrtl.WireVector).

- Convert indented comments to Markdown code blocks::

      #     def foo(bar):
      #         pass

  will be converted to::

      ```python
      def foo(bar):
          pass
      ```

Note the five spaces between `#` and `def`. We could instead write these code links and
code blocks directly in the source example scripts, but that would make the scripts
harder to read outside of a Jupyter notebook.

The transformation is very simple, and many edge cases are not handled. For example, any
line with a `#` character in column 0 is considered a comment, even if that line occurs
within a triple-quoted block.
"""

import argparse
import json

# Map from link anchor text to link target. This will be used to create hyperlinks for
# code references in the generated Jupyter notebook.
_link_map = {
    "Block": "https://pyrtl.readthedocs.io/en/latest/blocks.html#pyrtl.Block",
    "Const": "https://pyrtl.readthedocs.io/en/latest/basic.html#pyrtl.Const",
    "Input": "https://pyrtl.readthedocs.io/en/latest/basic.html#pyrtl.Input",
    "MemBlock": "https://pyrtl.readthedocs.io/en/latest/regmem.html#pyrtl.MemBlock",
    "Output": "https://pyrtl.readthedocs.io/en/latest/basic.html#pyrtl.Output",
    "Register": "https://pyrtl.readthedocs.io/en/latest/basic.html#pyrtl.Register",
    "RomBlock": "https://pyrtl.readthedocs.io/en/latest/regmem.html#pyrtl.RomBlock",
    "Simulation": "https://pyrtl.readthedocs.io/en/latest/simtest.html#pyrtl.Simulation",
    "SimulationTrace": "https://pyrtl.readthedocs.io/en/latest/simtest.html#pyrtl.SimulationTrace",
    "WireVector": "https://pyrtl.readthedocs.io/en/latest/basic.html#pyrtl.WireVector",
    "concat()": "https://pyrtl.readthedocs.io/en/latest/helpers.html#pyrtl.concat",
    "conditional_assignment": "https://pyrtl.readthedocs.io/en/latest/basic.html#conditional-assignment",
    "input_from_blif()": "https://pyrtl.readthedocs.io/en/latest/export.html#pyrtl.input_from_blif",
    "kogge_stone()": "https://pyrtl.readthedocs.io/en/latest/rtllib.html#pyrtl.rtllib.adders.kogge_stone",
    "mux()": "https://pyrtl.readthedocs.io/en/latest/helpers.html#pyrtl.mux",
    "otherwise": "https://pyrtl.readthedocs.io/en/latest/basic.html#pyrtl.otherwise",
    "output_to_svg()": "https://pyrtl.readthedocs.io/en/latest/export.html#pyrtl.block_to_svg",
    "output_to_trivialgraph()": "https://pyrtl.readthedocs.io/en/latest/export.html#pyrtl.output_to_trivialgraph",
    "output_verilog_testbench()": "https://pyrtl.readthedocs.io/en/latest/export.html#pyrtl.output_verilog_testbench",
    "probe()": "https://pyrtl.readthedocs.io/en/latest/helpers.html#pyrtl.probe",
    "render_trace()": "https://pyrtl.readthedocs.io/en/latest/simtest.html#pyrtl.SimulationTrace.render_trace",
    "reset_working_block()": "https://pyrtl.readthedocs.io/en/latest/blocks.html#pyrtl.reset_working_block",
    "set_debug_mode()": "https://pyrtl.readthedocs.io/en/latest/helpers.html#pyrtl.set_debug_mode",
    "sign_extended()": "https://pyrtl.readthedocs.io/en/latest/basic.html#pyrtl.WireVector.sign_extended",
    "signed_add()": "https://pyrtl.readthedocs.io/en/latest/helpers.html#pyrtl.signed_add",
    "signed_lt()": "https://pyrtl.readthedocs.io/en/latest/helpers.html#pyrtl.signed_lt",
    "signed_mult()": "https://pyrtl.readthedocs.io/en/latest/helpers.html#pyrtl.signed_mult",
    "signed_sub()": "https://pyrtl.readthedocs.io/en/latest/helpers.html#pyrtl.signed_sub",
    "step()": "https://pyrtl.readthedocs.io/en/latest/simtest.html#pyrtl.Simulation.step",
    "step_multiple()": "https://pyrtl.readthedocs.io/en/latest/simtest.html#pyrtl.Simulation.step_multiple",
    "synthesize()": "https://pyrtl.readthedocs.io/en/latest/analysis.html#pyrtl.synthesize",
    "tree_multiplier()": "https://pyrtl.readthedocs.io/en/latest/rtllib.html#pyrtl.rtllib.multipliers.tree_multiplier",
    "truncate()": "https://pyrtl.readthedocs.io/en/latest/basic.html#pyrtl.WireVector.truncate",
    "val_to_signed_integer()": "https://pyrtl.readthedocs.io/en/latest/helpers.html#pyrtl.val_to_signed_integer",
    "wire_matrix()": "https://pyrtl.readthedocs.io/en/latest/helpers.html#pyrtl.wire_matrix",
    "wire_struct": "https://pyrtl.readthedocs.io/en/latest/helpers.html#pyrtl.wire_struct",
    "working_block()": "https://pyrtl.readthedocs.io/en/latest/blocks.html#pyrtl.working_block",
    "zero_extended()": "https://pyrtl.readthedocs.io/en/latest/basic.html#pyrtl.WireVector.zero_extended",
}


# Convert _link_map to _replacement_map, which will be used for search-and-replace. This
# converts all the link anchor text that we search for to Markdown `code format`, and
# creates Markdown links for the replacement text. This also adds plurals to the
# _replacement_map, for example `WireVectors` will be linked to pyrtl.WireVector.
def _make_replacement_map(link_map: dict[str, str]) -> dict[str, str]:
    replacement_map = {}
    for anchor, target in link_map.items():
        replacement_map[f"`{anchor}`"] = f"[{anchor}]({target})"
        replacement_map[f"`{anchor}s`"] = f"[{anchor}s]({target})"
    return replacement_map


_replacement_map = _make_replacement_map(_link_map)

# Template for markdown cells. "source" will be added later, and it will contain the
# actual contents of the Markdown cell.
_markdown_template = {
    "cell_type": "markdown",
    "metadata": {},
}

# Template for code cells. "source" will be added later, and it will contain the actual
# contents of the code cell.
_code_template = {
    "cell_type": "code",
    "execution_count": None,
    "metadata": {"collapsed": True},
    "outputs": [],
}

# ipynb metadata.
_metadata = {
    "metadata": {
        "kernelspec": {
            "display_name": "Python 3",
            "language": "python",
            "name": "python3",
        },
        "language_info": {
            "codemirror_mode": {"name": "ipython", "version": 3},
            "file_extension": ".py",
            "mimetype": "text/x-python",
            "name": "python",
            "nbconvert_exporter": "python",
            "pygments_lexer": "ipython3",
            "version": "3.6.4",
        },
    },
    "nbformat": 4,
    "nbformat_minor": 2,
}


def to_ipynb(source_name: str, target_name: str):
    """Create a Jupyter notebook from a Python script.

    Comment blocks will be converted to markdown cells, and code blocks will be
    converted to code cells.

    :param source_name: Name of the Python script to convert.
    :param target_name: Name of the Jupyter notebook to generate.
    """
    # Collection of all markdown and code cells.
    cells = []
    # Current markdown cell contents. This will be the value for the markdown cell's
    # "source" key. If this is nonempty, `current_code_source` should be empty.
    current_markdown_source = []
    # Current code cell contents. This will be the valued for the code cell's "source"
    # key. If this is nonempty, `current_markdown_source` should be empty.
    current_code_source = []
    # Becomes True after emitting a code cell containing `import` statements.
    found_imports = False
    # Current code block within a markdown cell. Will be appended to
    # `current_markdown_source`, as a markdown code block.
    current_markdown_code_block = []

    def _emit_code():
        """Create a code cell for `current_code_source`, if necessary."""
        nonlocal current_code_source, found_imports
        if len(current_code_source) > 0:
            # Remove any trailing blank lines from the code cell.
            while current_code_source[-1].rstrip() == "":
                current_code_source = current_code_source[:-1]

                for code_line in current_code_source:
                    if not found_imports and (
                        code_line.startswith("import pyrtl")
                        or code_line.startswith("from pyrtl")
                    ):
                        found_imports = True
                        pip_packages = "pyrtl"
                        if "verilog" in source_name:
                            pip_packages = "pyrtl pyparsing"

                        current_code_source = [
                            f"%pip install {pip_packages}\n",
                            "\n",
                            *current_code_source,
                            "\n",
                            "pyrtl.reset_working_block()\n",
                        ]
                        break

            cells.append(_code_template | {"source": current_code_source})
            current_code_source = []

    def _emit_markdown_code_block():
        """Add a code block within `current_markdown_source`, if necessary."""
        nonlocal current_markdown_code_block
        if len(current_markdown_code_block) > 0:
            current_markdown_source.extend(
                ["```python\n", *current_markdown_code_block, "```\n"]
            )
            current_markdown_code_block = []

    def _emit_markdown():
        """Create a markdown cell for `current_markdown_source`, if necessary."""
        nonlocal current_markdown_source
        _emit_markdown_code_block()
        if len(current_markdown_source) > 0:
            cells.append(_markdown_template | {"source": current_markdown_source})
            current_markdown_source = []

    with open(source_name) as source_file:
        for current_line in source_file:
            if current_line.startswith("#"):
                # `current_line` is a comment. Append contiguous comment lines to
                # `current_markdown_source` or `current_markdown_code_block`.
                _emit_code()
                # Drop the "#" at the start of the line.
                current_line = current_line[1:]

                if current_line.startswith("     "):
                    # Current line starts or extends a code block.
                    current_markdown_code_block.append(current_line[5:])
                else:
                    _emit_markdown_code_block()
                    for old, new in _replacement_map.items():
                        current_line = current_line.replace(old, new)
                    current_markdown_source.append(current_line)

            elif current_line.rstrip() == "":
                if len(current_markdown_code_block) > 0:
                    # In a code block, a blank line extends the code block.
                    current_markdown_code_block.append(current_line)
                elif len(current_markdown_source) > 0:
                    # In a markdown cell, a blank line ends the markdown cell.
                    _emit_markdown()
                elif len(current_code_source) > 0:
                    # In a code cell, a blank line extends the code cell.
                    current_code_source.append(current_line)

            else:
                # Else `current_line` is code. Add it to `current_code_source`.
                _emit_markdown()
                current_code_source.append(current_line)

    # Emit any remaining cell data.
    _emit_markdown()
    _emit_code()

    # Generate the .ipynb file.
    ipynb = {"cells": cells} | _metadata
    with open(target_name, "w") as target_file:
        json.dump(ipynb, target_file, indent=1)


def main():
    parser = argparse.ArgumentParser(prog="to_ipynb.py")
    parser.add_argument("source_name", type=str)
    parser.add_argument("target_name", type=str)
    args = parser.parse_args()

    to_ipynb(args.source_name, args.target_name)


if __name__ == "__main__":
    main()



import doctest
import random
import unittest

import pyrtl
import pyrtl.rtllib.testingutils as utils
from pyrtl.rtllib import adders


class TestDocTests(unittest.TestCase):
    """Test documentation examples."""

    def test_doctests(self):
        failures, tests = doctest.testmod(m=pyrtl.rtllib.adders)
        self.assertGreater(tests, 0)
        self.assertEqual(failures, 0)


class TestAdders(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        random.seed(8492049)

    def setUp(self):
        pyrtl.reset_working_block()

    def tearDown(self):
        pyrtl.reset_working_block()

    def adder2_t_base_1(self, adder_func):
        self.adder_t_base(adder_func, max_bitwidth=34, num_wires=2)

    def adder_t_base(self, adder_func, **kwargs):
        wires, vals = utils.make_inputs_and_values(
            dist=utils.inverse_power_dist, **kwargs
        )
        outwire = pyrtl.Output(name="test")
        outwire <<= adder_func(*wires)

        out_vals = utils.sim_and_ret_out(outwire, wires, vals)
        true_result = [sum(cycle_vals) for cycle_vals in zip(*vals)]
        self.assertEqual(out_vals, true_result)

    def test_kogge_stone_1(self):
        self.adder2_t_base_1(adders.kogge_stone)

    def test_ripple_1(self):
        self.adder2_t_base_1(adders.ripple_add)

    def test_carrylookahead_1(self):
        self.adder2_t_base_1(adders.cla_adder)

    def test_carry_save_1(self):
        self.adder_t_base(adders.carrysave_adder, exact_bitwidth=32, num_wires=3)

    def test_fast_group_adder_1(self):
        wires, vals = utils.make_inputs_and_values(
            max_bitwidth=12, num_wires=7, dist=utils.inverse_power_dist
        )
        outwire = pyrtl.Output(name="test")
        outwire <<= adders.fast_group_adder(wires)

        out_vals = utils.sim_and_ret_out(outwire, wires, vals)
        true_result = [sum(cycle_vals) for cycle_vals in zip(*vals)]
        self.assertEqual(out_vals, true_result)


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

import builtins
from functools import reduce

from pyrtl import (
    Const,
    PyrtlError,
    WireVector,
    as_wires,
    concat,
    formatted_str_to_val,
    select,
)
from pyrtl.rtllib.multipliers import fused_multiply_adder
from pyrtl.wire import WireVectorLike


class Matrix:
    """Class for making a Matrix using PyRTL.

    Provides the ability to perform different matrix operations.
    """

    # Internally, this class uses a Python list of lists of WireVectors.
    # So, a Matrix is represented as follows for a 2 x 2:
    # [[WireVector, WireVector],
    #  [WireVector, WireVector]]
    def __init__(
        self,
        rows: int,
        columns: int,
        bits: int,
        signed: bool = False,  # noqa: ARG002
        value: WireVector | list[list[WireVectorLike]] = None,
        max_bits: int = 64,
    ):
        """Constructs a Matrix object.

        :param rows: The number of rows in the matrix. Must be greater than 0.
        :param columns: The number of columns in the matrix. Must be greater than 0.
        :param bits: The number of bits per :class:`.WireVector` matrix element. Must be
            greater than 0.
        :param signed: Currently not supported (will be added in the future).
        :param value: The value you want to initialize the ``Matrix`` to. If a
            :class:`.WireVector`, must be of size ``rows * columns * bits``. If a
            :class:`list`, must have ``rows`` rows and ``columns`` columns, and every
            element must be representable with a :attr:`~.WireVector.bitwidth` of
            ``bits``. If ``None``, the matrix initializes to 0.
        :param max_bits: The maximum number of bits each :class:`.WireVector` element
            can grow to. Operations like multiplication and addition can produce
            matrices with more ``bits``, but results will be limited to ``max_bits``.
        """
        if not isinstance(rows, int):
            msg = (
                f'Rows must be of type int, instead "{rows}" was passed of type '
                f"{type(rows)}"
            )
            raise PyrtlError(msg)
        if rows <= 0:
            msg = (
                f"Rows cannot be less than or equal to zero. Rows value passed: {rows}"
            )
            raise PyrtlError(msg)

        if not isinstance(columns, int):
            msg = (
                f'Columns must be of type int, instead "{columns}" was passed of type '
                f"{type(columns)}"
            )
            raise PyrtlError(msg)
        if columns <= 0:
            msg = (
                "Columns cannot be less than or equal to zero. Columns value passed: "
                f"{columns}"
            )
            raise PyrtlError(msg)

        if not isinstance(bits, int):
            msg = (
                f'Bits must be of type int, instead "{bits}" was passed of type '
                f"{type(bits)}"
            )
            raise PyrtlError(msg)
        if bits <= 0:
            msg = f'Bits cannot be negative or zero, instead "{bits}" was passed'
            raise PyrtlError(msg)

        if max_bits is not None and bits > max_bits:
            bits = max_bits

        self._matrix = [[0 for _ in range(columns)] for _ in range(rows)]

        if value is None:
            for i in range(rows):
                for j in range(columns):
                    self._matrix[i][j] = Const(0)
        elif isinstance(value, WireVector):
            if value.bitwidth != bits * rows * columns:
                msg = (
                    "Initialized bitwidth value does not match given value.bitwidth: "
                    f"{value.bitwidth}, expected: {bits * rows * columns}"
                )
                raise PyrtlError(msg)
            for i in range(rows):
                for j in range(columns):
                    start_index = (j * bits) + (i * columns * bits)
                    self._matrix[rows - i - 1][columns - j - 1] = as_wires(
                        value[start_index : start_index + bits], bitwidth=bits
                    )

        elif isinstance(value, list):
            if len(value) != rows or any(len(row) != columns for row in value):
                msg = (
                    "Rows and columns mismatch\n"
                    f"Rows: {len(value)}, expected: {rows}\n"
                    f"Columns: {len(value[0])}, expected: {columns}"
                )
                raise PyrtlError(msg)
            for i in range(rows):
                for j in range(columns):
                    self._matrix[i][j] = as_wires(value[i][j], bitwidth=bits)

        else:
            msg = (
                "Initialized value must be of type WireVector or list. Instead was "
                f"passed value of type {type(value)}"
            )
            raise PyrtlError(msg)

        self.rows = rows
        self.columns = columns
        self._bits = bits
        self.bits = bits
        self.signed = False
        self.max_bits = max_bits

    @property
    def bits(self) -> int:
        """The number of bits for each matrix element.

        Reducing the number of ``bits`` will :meth:`~.WireVector.truncate` the most
        significant bits of each matrix element.
        """
        return self._bits

    @bits.setter
    def bits(self, bits):
        if not isinstance(bits, int):
            msg = (
                f'Bits must be of type int, instead "{bits}" was passed of type '
                f"{type(bits)}"
            )
            raise PyrtlError(msg)
        if bits <= 0:
            msg = f'Bits cannot be negative or zero, instead "{bits}" was passed'
            raise PyrtlError(msg)

        self._bits = bits
        for i in range(self.rows):
            for j in range(self.columns):
                self._matrix[i][j] = self._matrix[i][j][:bits]

    def __len__(self) -> int:
        """Returns the total bitwidth for all elements in the ``Matrix``.

        :return: The ``Matrix``'s total :attr:`~.WireVector.bitwidth`: ``rows * columns
            * bits``.
        """
        return self.bits * self.rows * self.columns

    def to_wirevector(self) -> WireVector:
        """Returns all elements in the ``Matrix`` in one :class:`.WireVector`.

        This :func:`concatenates<.concat>` all the ``Matrix``'s elements together, in
        row-major order.

        For example, a 2 x 1 matrix ``[[wire_a, wire_b]]`` would become
        ``pyrtl.concat(wire_a, wire_b)``.

        :return: A concatenated :class:`.WireVector` containing all of the ``Matrix``'s
                 elements.
        """
        result = []

        for i in range(len(self._matrix)):
            for j in range(len(self._matrix[0])):
                result.append(as_wires(self[i, j], bitwidth=self.bits))

        return as_wires(concat(*result), bitwidth=len(self))

    def transpose(self) -> Matrix:
        """
        :return: A ``Matrix`` representing the transpose of ``self``.
        """
        result = Matrix(self.columns, self.rows, self.bits, max_bits=self.max_bits)
        for i in range(result.rows):
            for j in range(result.columns):
                result[i, j] = self[j, i]
        return result

    def __reversed__(self) -> Matrix:
        """Invoked with the :func:`reversed` builtin.

        :return: A ``Matrix`` with all row and column indices reversed.
        """
        result = Matrix(self.rows, self.columns, self.bits, max_bits=self.max_bits)
        for i in range(self.rows):
            for j in range(self.columns):
                result[i, j] = self[self.rows - 1 - i, self.columns - 1 - j]
        return result

    def __getitem__(self, key: int | slice | tuple[int, int]) -> WireVector | Matrix:
        """Access elements in the ``Matrix``.

        Invoked with square brackets, like ``matrix[...]``.

        Examples::

            int_matrix = [[0, 1, 2], [3, 4, 5], [6, 7, 8]]
            matrix = Matrix(rows=3, columns=3, bits=4, value=int_matrix)

            # Retrieve the second row.
            matrix[1] == [3, 4, 5]

            # Retrieve the last row.
            matrix[-1] == [6, 7, 8]

            # Retrieve the element in row 2, column 0.
            matrix[2, 0] == 6
            matrix[(2, 0)] = 6

            # Retrieve the first two rows.
            matrix[slice(0, 2), slice(0, 3)] == [[0, 1, 2], [3, 4, 5]]
            matrix[0:2, 0:3] == [[0, 1, 2], [3, 4, 5]]
            matrix[:2] == [[0, 1, 2], [3, 4, 5]]

            # Retrieve the last two rows.
            matrix[-2:] == [[3, 4, 5], [6, 7, 8]]

        :param key: The key value to get.

        :return: :class:`.WireVector` or ``Matrix`` containing the value of key.
        """
        if isinstance(key, tuple):
            rows, columns = key

            # First set up proper slice
            if not isinstance(rows, slice):
                if not isinstance(rows, int):
                    msg = (
                        f'Rows must be of type int or slice, instead "{rows}" was '
                        f"passed of type {type(rows)}"
                    )
                    raise PyrtlError(msg)
                if rows < 0:
                    rows = self.rows - abs(rows)
                    if rows < 0:
                        msg = (
                            f"Invalid bounds for rows. Max rows: {self.rows}, got: "
                            f"{rows}"
                        )
                        raise PyrtlError(msg)
                rows = slice(rows, rows + 1, 1)

            if not isinstance(columns, slice):
                if not isinstance(columns, int):
                    msg = (
                        f'Columns must be of type int or slice, instead "{columns}" '
                        f"was passed of type {type(columns)}"
                    )
                    raise PyrtlError(msg)
                if columns < 0:
                    columns = self.columns - abs(columns)
                    if columns < 0:
                        msg = (
                            f"Invalid bounds for columns. Max columns: {self.columns}, "
                            f"got: {columns}"
                        )
                        raise PyrtlError(msg)
                columns = slice(columns, columns + 1, 1)

            if rows.start is None:
                rows = slice(0, rows.stop, rows.step)
            elif rows.start < 0:
                rows = slice(self.rows - abs(rows.start), rows.stop, rows.step)

            if rows.stop is None:
                rows = slice(rows.start, self.rows, rows.step)
            elif rows.stop < 0:
                rows = slice(rows.start, self.rows - abs(rows.stop), rows.step)

            rows = slice(rows.start, rows.stop, 1)

            if columns.start is None:
                columns = slice(0, columns.stop, columns.step)
            elif columns.start < 0:
                columns = slice(
                    self.columns - abs(columns.start), columns.stop, columns.step
                )

            if columns.stop is None:
                columns = slice(columns.start, self.columns, columns.step)
            elif columns.stop < 0:
                columns = slice(
                    columns.start, self.columns - abs(columns.stop), columns.step
                )

            columns = slice(columns.start, columns.stop, 1)

            # Check slice bounds
            if (
                rows.start > self.rows
                or rows.stop > self.rows
                or rows.start < 0
                or rows.stop < 0
            ):
                msg = (
                    f"Invalid bounds for rows. Max rows: {self.rows}, got: "
                    f"{rows.start}:{rows.stop}"
                )
                raise PyrtlError(msg)
            if (
                columns.start > self.columns
                or columns.stop > self.columns
                or columns.start < 0
                or columns.stop < 0
            ):
                msg = (
                    f"Invalid bounds for columns. Max columns: {self.columns}, got: "
                    f"{columns.start}:{columns.stop}"
                )
                raise PyrtlError(msg)

            # If it's a single value we want to return a wirevector
            if rows.stop - rows.start == 1 and columns.stop - columns.start == 1:
                return as_wires(self._matrix[rows][0][columns][0], bitwidth=self.bits)

            # Otherwise set up matrix and return that
            result = [
                [0 for _ in range(columns.stop - columns.start)]
                for _ in range(rows.stop - rows.start)
            ]

            for i in range(len(result)):
                for j in range(len(result[0])):
                    result[i][j] = self._matrix[i + rows.start][j + columns.start]
            return Matrix(
                len(result),
                len(result[0]),
                self._bits,
                signed=self.signed,
                value=result,
                max_bits=self.max_bits,
            )

        # Second case when we just want to get full row
        if isinstance(key, int):
            if key < 0:
                start = self.rows - abs(key)
                if start < 0:
                    msg = (
                        f"Index {key} is out of bounds for matrix with {self.rows} rows"
                    )
                    raise PyrtlError(msg)
                key = slice(start, start + 1, None)
            else:
                key = slice(key, key + 1, None)
            return self[key, :]

        # Third case when we want multiple rows
        if isinstance(key, slice):
            return self[key, :]

        # Otherwise improper value was passed
        msg = (
            f'Rows must be of type int or slice, instead "{key}" was passed of type '
            f"{type(key)}"
        )
        raise PyrtlError(msg)

    def __setitem__(
        self,
        key: int | slice | tuple[int, int],
        value: WireVectorLike | Matrix,
    ):
        """Mutate the ``Matrix``.

        Invoked with square brackets, like ``matrix[a, b] = value``. ``value`` will be
        truncated so it fits in :attr:`bits`.

        This modifies the ``Matrix``'s :class:`lists<list>` in its internal
        :class:`list` of :class:`list` of :class:`WireVectors<.WireVector>`, which makes
        the ``Matrix`` use a different set of :class:`WireVectors<.WireVector>` as its
        elements. It does not modify any :class:`WireVectors<.WireVector>`.

        :param key: The key value to set.
        :param value: The value in which to set the key.
        """

        if isinstance(key, tuple):
            rows, columns = key

            # First ensure that slices are correct
            if not isinstance(rows, slice):
                if not isinstance(rows, int):
                    msg = (
                        f'Rows must be of type int or slice, instead "{rows}" was '
                        f"passed of type {type(rows)}"
                    )
                    raise PyrtlError(msg)
                rows = slice(rows, rows + 1, 1)

            if not isinstance(columns, slice):
                if not isinstance(columns, int):
                    msg = (
                        f'Columns must be of type int or slice, instead "{columns}" '
                        f"was passed of type {type(columns)}"
                    )
                    raise PyrtlError(msg)
                columns = slice(columns, columns + 1, 1)

            if rows.start is None:
                rows = slice(0, rows.stop, rows.step)
            elif rows.start < 0:
                rows = slice(self.rows - abs(rows.start), rows.stop, rows.step)

            if rows.stop is None:
                rows = slice(rows.start, self.rows, rows.step)
            elif rows.stop < 0:
                rows = slice(rows.start, self.rows - abs(rows.stop), rows.step)

            if columns.start is None:
                columns = slice(0, columns.stop, columns.step)
            elif columns.start < 0:
                columns = slice(
                    self.columns - abs(columns.start), columns.stop, columns.step
                )

            if columns.stop is None:
                columns = slice(columns.start, self.columns, columns.step)
            elif columns.stop < 0:
                columns = slice(
                    columns.start, self.columns - abs(columns.stop), columns.step
                )

            # Check Slice Bounds
            if (
                rows.start > self.rows
                or rows.stop > self.rows
                or rows.start < 0
                or rows.stop < 0
            ):
                msg = (
                    f"Invalid bounds for rows. Max rows: {self.rows}, got: "
                    f"{rows.start}:{rows.stop}"
                )
                raise PyrtlError(msg)
            if (
                columns.start > self.columns
                or columns.stop > self.columns
                or columns.start < 0
                or columns.stop < 0
            ):
                msg = (
                    f"Invalid bounds for columns. Max columns: {self.columns}, got: "
                    f"{columns.start}:{columns.stop}"
                )
                raise PyrtlError(msg)

            # First case when setting value to Matrix
            if isinstance(value, Matrix):
                if value.rows != (rows.stop - rows.start):
                    msg = (
                        "Value rows mismatch. Expected Matrix of rows "
                        f'"{rows.stop - rows.start}", instead received Matrix of rows '
                        f'"{value.rows}"'
                    )
                    raise PyrtlError(msg)
                if value.columns != (columns.stop - columns.start):
                    msg = (
                        "Value columns mismatch. Expected Matrix of columns "
                        f'"{columns.stop - columns.start}", instead received Matrix of '
                        f'columns "{value.columns}"'
                    )
                    raise PyrtlError(msg)

                for i in range(rows.stop - rows.start):
                    for j in range(columns.stop - columns.start):
                        self._matrix[rows.start + i][columns.start + j] = as_wires(
                            value[i, j], bitwidth=self.bits
                        )

            # Second case when setting value to wirevector
            elif isinstance(value, (int, WireVector)):
                if ((rows.stop - rows.start) != 1) or (
                    (columns.stop - columns.start) != 1
                ):
                    msg = "Value mismatch: expected Matrix, instead received WireVector"
                    raise PyrtlError(msg)
                self._matrix[rows.start][columns.start] = as_wires(
                    value, bitwidth=self.bits
                )
            # Otherwise Error
            else:
                msg = f"Invalid value of type {type(value)}"
                raise PyrtlError(msg)
        else:
            # Second case if we just want to set a full row
            if isinstance(key, int):
                if key < 0:
                    start = self.rows - abs(key)
                    if start < 0:
                        msg = (
                            f"Index {key} is out of bounds for matrix with {self.rows} "
                            "rows"
                        )
                        raise PyrtlError(msg)
                    key = slice(start, start + 1, None)
                else:
                    key = slice(key, key + 1, None)
                self[key, :] = value
            # Third case if we want to set full rows
            elif isinstance(key, slice):
                self[key, :] = value
            else:
                msg = (
                    f'Rows must be of type int or slice, instead "{key}" was passed of '
                    f"type {type(key)}"
                )
                raise PyrtlError(msg)

    def copy(self) -> Matrix:
        """Constructs a copy of the ``Matrix``.

        The returned copy will have new set of :class:`WireVectors<.WireVector>` for its
        elements, but each new :class:`WireVector` will be wired to the corresponding
        :class:`WireVector` in the original ``Matrix``.

        :return: A new instance of ``Matrix`` that indirectly refers to the same
                 underlying :class:`WireVectors<.WireVector>` as ``self``.
        """
        return Matrix(
            self.rows,
            self.columns,
            self.bits,
            value=self.to_wirevector(),
            max_bits=self.max_bits,
        )

    def __iadd__(self, other: Matrix) -> Matrix:
        """Perform the in-place addition operation.

        Invoked with ``a += b``. Performs elementwise addition.

        :return: a Matrix object with the elementwise addition being performed.
        """
        new_value = self + other
        self._matrix = new_value._matrix
        self.bits = new_value._bits
        return self.copy()

    def __add__(self, other: Matrix) -> Matrix:
        """Perform the addition operation.

        Invoked with ``a + b``. Performs elementwise addition.

        :return: a Matrix object containing the elementwise sum.
        """
        if not isinstance(other, Matrix):
            msg = f"error: expecting a Matrix, got {type(other)} instead"
            raise PyrtlError(msg)

        if self.columns != other.columns:
            msg = (
                f"error: columns mismatch. Matrix a: {self.columns} columns, Matrix b: "
                f"{other.columns} columns"
            )
            raise PyrtlError(msg)
        if self.rows != other.rows:
            msg = (
                f"error: row mismatch. Matrix a: {self.rows} rows, Matrix b: "
                f"{other.rows} rows"
            )
            raise PyrtlError(msg)

        new_bits = self.bits
        if other.bits > new_bits:
            new_bits = other.bits

        result = Matrix(self.rows, self.columns, new_bits + 1, max_bits=self.max_bits)

        for i in range(result.rows):
            for j in range(result.columns):
                result[i, j] = self[i, j] + other[i, j]
        return result

    def __isub__(self, other: Matrix) -> Matrix:
        """Perform the inplace subtraction opperation.

        Invoked with ``a -= b``. Performs elementwise subtraction.

        :param other: The ``Matrix`` to subtract.

        :return: A ``Matrix`` object with the result of elementwise subtraction.
        """
        new_value = self - other
        self._matrix = new_value._matrix
        self._bits = new_value._bits
        return self.copy()

    def __sub__(self, other: Matrix) -> Matrix:
        """Perform the subtraction operation.

        Invoked with ``a - b``. Performs elementwise subtraction.

        .. note::

            If ``signed=False``, the result will be floored at 0.

        :param other: The ``Matrix`` to subtract.

        :return: a ``Matrix`` object with the result of elementwise subtraction.
        """
        if not isinstance(other, Matrix):
            msg = f"error: expecting a Matrix, got {type(other)} instead"
            raise PyrtlError(msg)

        if self.columns != other.columns:
            msg = (
                f"error: columns mismatch. Matrix a: {self.columns} columns, "
                f"Matrix b: {other.columns} columns"
            )
            raise PyrtlError(msg)

        if self.rows != other.rows:
            msg = (
                f"error: row mismatch. Matrix a: {self.rows} rows, Matrix b: "
                f"{other.rows} rows"
            )
            raise PyrtlError(msg)

        new_bits = self.bits
        if other.bits > new_bits:
            new_bits = other.bits

        result = Matrix(self.rows, self.columns, new_bits, max_bits=self.max_bits)

        for i in range(result.rows):
            for j in range(result.columns):
                if self.signed:
                    result[i, j] = self[i, j] - other[i, j]
                else:
                    result[i, j] = select(
                        self[i, j] > other[i, j], self[i, j] - other[i, j], Const(0)
                    )

        return result

    def __imul__(self, other: Matrix | WireVector) -> Matrix:
        """Perform the in-place multiplication operation.

        Invoked with ``a *= b``. Performs elementwise or scalar multiplication.

        :param other: The ``Matrix`` or scalar to multiply.

        :return: A ``Matrix`` object with the product.
        """
        new_value = self * other
        self._matrix = new_value._matrix
        self._bits = new_value._bits
        return self.copy()

    def __mul__(self, other: Matrix | WireVector) -> Matrix:
        """Perform the elementwise or scalar multiplication operation.

        Invoked with ``a * b``.

        :param other: The ``Matrix`` or scalar to multiply.

        :return: A ``Matrix`` object with the product.
        """

        if isinstance(other, Matrix):
            if self.columns != other.columns:
                msg = (
                    f"error: columns mismatch. Matrix a: {self.columns} columns, "
                    f"Matrix b: {other.columns} columns"
                )
                raise PyrtlError(msg)
            if self.rows != other.rows:
                msg = (
                    f"error, row mismatch Matrix a: {self.rows} rows, Matrix b: "
                    f"{other.rows} rows"
                )
                raise PyrtlError(msg)

            bits = self.bits + other.bits
        elif isinstance(other, WireVector):
            bits = self.bits + len(other)
        else:
            msg = f"Expecting a Matrix or WireVector got {type(other)} instead"
            raise PyrtlError(msg)

        result = Matrix(self.rows, self.columns, bits, max_bits=self.max_bits)

        for i in range(self.rows):
            for j in range(self.columns):
                if isinstance(other, Matrix):
                    result[i, j] = self[i, j] * other[i, j]
                else:
                    result[i, j] = self[i, j] * other
        return result

    def __imatmul__(self, other: Matrix) -> Matrix:
        """Performs the inplace matrix multiplication operation.

        Invoked with ``a @= b``.

        :param other: The second ``Matrix``.

        :return: A ``Matrix`` that contains the product.
        """
        new_value = self.__matmul__(other)
        self.columns = new_value.columns
        self.rows = new_value.rows
        self._matrix = new_value._matrix
        self._bits = new_value._bits
        return self.copy()

    def __matmul__(self, other: Matrix) -> Matrix:
        """Performs the inplace matrix multiplication operation.

        Invoked with ``a @ b``.

        :param other: The second ``Matrix``.

        :return: A ``Matrix`` that contains the product.
        """
        if not isinstance(other, Matrix):
            msg = f"error: expecting a Matrix, got {type(other)} instead"
            raise PyrtlError(msg)

        if self.columns != other.rows:
            msg = (
                f"error: rows and columns mismatch. Matrix a: {self.columns} columns, "
                f"Matrix b: {other.rows} rows"
            )
            raise PyrtlError(msg)

        result = Matrix(
            self.rows,
            other.columns,
            self.columns * other.rows * (self.bits + other.bits),
            max_bits=self.max_bits,
        )

        for i in range(self.rows):
            for j in range(other.columns):
                for k in range(self.columns):
                    result[i, j] = fused_multiply_adder(
                        self[i, k], other[k, j], result[i, j], signed=self.signed
                    )

        return result

    def __ipow__(self, power: int) -> Matrix:
        """Performs the matrix power operation.

        Invoked with ``a **= b``.

        This performs a chain of matrix multiplications, where ``self`` is matrix
        multiplied by ``self``, ``power`` times.

        :param power: The power to raise the matrix to.

        :return: A ``Matrix`` containing the result.
        """
        new_value = self**power
        self._matrix = new_value._matrix
        self._bits = new_value._bits
        return self.copy()

    def __pow__(self, power: int) -> Matrix:
        """Performs the matrix power operation.

        Invoked with ``a ** b``.

        This performs a chain of matrix multiplications, where ``self`` is matrix
        multiplied by ``self``, ``power`` times.

        :param power: The power to raise the matrix to.

        :return: A ``Matrix`` containing the result.
        """
        if not isinstance(power, int):
            msg = (
                "Unexpected power given. Type int expected, but received type "
                f"{type(power)}"
            )
            raise PyrtlError(msg)

        if self.rows != self.columns:
            msg = "Matrix must be square"
            raise PyrtlError(msg)

        result = self.copy()

        # First case: return identity matrix
        if power == 0:
            for i in range(self.rows):
                for j in range(self.columns):
                    if i != j:
                        result[i, j] = Const(0)
                    else:
                        result[i, j] = Const(1)
            return result
        # Second case: do matrix multiplications
        if power >= 1:
            inputs = [result] * power

            def pow_2(first, second):
                return first.__matmul__(second)

            return reduce(pow_2, inputs)

        msg = "Power must be greater than or equal to 0"
        raise PyrtlError(msg)

    def put(
        self,
        ind: int | list[int] | tuple[int],
        v: int | list[int] | tuple[int] | Matrix,
        mode: str = "raise",
    ):
        """Replace specified elements of the ``Matrix`` with values ``v``.

        Note that the index ``ind`` is on the flattened matrix.

        :param ind: Target indices.
        :param v: Values to place in ``Matrix`` at ``ind``. If ``v`` is shorter than
                  ``ind``, ``v`` will be repeated.
        :param mode: How out-of-bounds indices behave. ``raise`` raises an error,
            ``wrap`` wraps around, and ``clip`` clips to the range.
        """
        count = self.rows * self.columns
        if isinstance(ind, int):
            ind = (ind,)
        elif not isinstance(ind, (tuple, list)):
            msg = f"Expected int or list-like indices, got {type(ind)}"
            raise PyrtlError(msg)

        if isinstance(v, int):
            v = (v,)

        if isinstance(v, (tuple, list)) and len(v) == 0:
            return
        if isinstance(v, Matrix) and v.rows != 1:
            msg = f"Expected a row-vector matrix, instead got matrix with {v.rows} rows"
            raise PyrtlError(msg)

        if mode not in ["raise", "wrap", "clip"]:
            msg = (
                f"Unexpected mode {mode}; allowable modes are 'raise', 'wrap', and "
                "'clip'"
            )
            raise PyrtlError(msg)

        def get_ix(ix):
            if ix < 0:
                ix = count - abs(ix)
            if ix < 0 or ix >= count:
                if mode == "raise":
                    msg = f"index {ix} is out of bounds with size {count}"
                    raise PyrtlError(msg)
                if mode == "wrap":
                    ix = ix % count
                elif mode == "clip":
                    ix = 0 if ix < 0 else count - 1
            return ix

        def get_value(ix):
            if isinstance(v, (tuple, list)):
                if ix >= len(v):
                    return v[-1]  # if v is shorter than ind, repeat last as necessary
                return v[ix]
            if isinstance(v, Matrix):
                if ix >= count:
                    return v[0, -1]
                return v[0, ix]
            return None

        for v_ix, mat_ix in enumerate(ind):
            mat_ix = get_ix(mat_ix)
            row = mat_ix // self.columns
            col = mat_ix % self.columns
            self[row, col] = get_value(v_ix)

    def reshape(self, *newshape: int | tuple, order: str = "C"):
        """Create a ``Matrix`` of the given shape from ``self``.

        One shape dimension in ``newshape`` can be ``-1``; in this case, the value for
        that dimension is inferred from the other given dimension (if any) and the
        number of elements in ``self``.

        Examples::

            int_matrix = [[0, 1, 2, 3], [4, 5, 6, 7]]
            matrix = Matrix.Matrix(2, 4, 4, value=int_matrix)

            matrix.reshape(-1) == [[0, 1, 2, 3, 4, 5, 6, 7]]
            matrix.reshape(8) == [[0, 1, 2, 3, 4, 5, 6, 7]]
            matrix.reshape(1, 8) == [[0, 1, 2, 3, 4, 5, 6, 7]]
            matrix.reshape((1, 8)) == [[0, 1, 2, 3, 4, 5, 6, 7]]
            matrix.reshape((1, -1)) == [[0, 1, 2, 3, 4, 5, 6, 7]]

            matrix.reshape(4, 2) == [[0, 1], [2, 3], [4, 5], [6, 7]]
            matrix.reshape(-1, 2) == [[0, 1], [2, 3], [4, 5], [6, 7]]
            matrix.reshape(4, -1) == [[0, 1], [2, 3], [4, 5], [6, 7]]

        :param newshape: Shape of the matrix to return. If ``newshape`` is a single
            :class:`int`, the new shape will be a 1-D row-vector of that length. If
            ``newshape`` is a :class:`tuple`, the :class:`tuple` specifies the new
            number of rows and columns. ``newshape`` can also be varargs.
        :param order: ``C`` means to read from self using row-major order (C-style), and
            ``F`` means to read from self using column-major order (Fortran-style).

        :return: A copy of the matrix with same data, with a new number of rows and
                 columns.
        """
        count = self.rows * self.columns
        if isinstance(newshape, int):
            if newshape == -1:
                newshape = (1, count)
            else:
                newshape = (1, newshape)
        elif isinstance(newshape, tuple):
            if isinstance(newshape[0], tuple):
                newshape = newshape[0]
            if len(newshape) == 1:
                newshape = (1, newshape[0])
            if len(newshape) > 2:
                msg = "length of newshape tuple must be <= 2"
                raise PyrtlError(msg)
            rows, cols = newshape
            if not isinstance(rows, int) or not isinstance(cols, int):
                msg = (
                    "newshape dimensions must be integers, instead got "
                    f"{type(newshape)}"
                )
                raise PyrtlError(msg)
            if rows == -1 and cols == -1:
                msg = "Both dimensions in newshape cannot be -1"
                raise PyrtlError(msg)
            if rows == -1:
                rows = count // cols
                newshape = (rows, cols)
            elif cols == -1:
                cols = count // rows
                newshape = (rows, cols)
        else:
            msg = (
                f"newshape can be an integer or tuple of integers, not {type(newshape)}"
            )
            raise PyrtlError(msg)

        rows, cols = newshape
        if rows * cols != count:
            msg = f"Cannot reshape matrix of size {count} into shape {newshape}"
            raise PyrtlError(msg)

        if order not in "CF":
            msg = (
                f"Invalid order {order}. Acceptable orders are 'C' (for row-major "
                "C-style order) and 'F' (for column-major Fortran-style order)."
            )
            raise PyrtlError(msg)

        value = [[0] * cols for _ in range(rows)]
        ix = 0
        if order == "C":
            # Read and write in row-wise order
            for newr in range(rows):
                for newc in range(cols):
                    r = ix // self.columns
                    c = ix % self.columns
                    value[newr][newc] = self[r, c]
                    ix += 1
        else:
            # Read and write in column-wise order
            for newc in range(cols):
                for newr in range(rows):
                    r = ix % self.rows
                    c = ix // self.rows
                    value[newr][newc] = self[r, c]
                    ix += 1

        return Matrix(rows, cols, self.bits, self.signed, value, self.max_bits)

    def flatten(self, order: str = "C"):
        """Flatten the ``Matrix`` into a single row.

        :param order: ``C`` means row-major order (C-style), and ``F`` means
            column-major order (Fortran-style)

        :return: A copy of the ``Matrix`` flattened into a row vector.
        """
        return self.reshape(self.rows * self.columns, order=order)


def multiply(first: Matrix, second: Matrix | WireVector) -> Matrix:
    """Perform the elementwise or scalar multiplication operation.

    .. WARNING::

        Use :meth:`Matrix.__mul__` instead.

    :param first: first matrix
    :param second: second matrix

    :return: a Matrix object with the element wise or scalar multiplication being
             performed
    """
    if not isinstance(first, Matrix):
        msg = f"error: expecting a Matrix, got {type(second)} instead"
        raise PyrtlError(msg)
    return first * second


def sum(
    matrix: Matrix | WireVector, axis: int | None = None, bits: int | None = None
) -> Matrix | WireVector:
    """Returns the sum of values in a ``Matrix`` across ``axis``.

    This performs a reduction, summing over the specified ``axis``.

    :param matrix: The matrix to perform sum operation on. If it is a
        :class:`.WireVector`, it will return itself.
    :param axis: The axis to perform the operation on. ``None`` refers to sum of all
        elements. ``0`` is sum of column. ``1`` is sum of rows. Defaults to ``None``.
    :param bits: The bits per element of the sum. Defaults to ``matrix.bits``.

    :return: A :class:`.WireVector` or ``Matrix`` representing the sum.
    """

    def sum_2(first, second):
        return first + second

    if isinstance(matrix, WireVector):
        return matrix

    if not isinstance(matrix, Matrix):
        msg = (
            f"error: expecting a Matrix or WireVector for matrix, got {type(matrix)} "
            "instead"
        )
        raise PyrtlError(msg)

    if not isinstance(bits, int) and bits is not None:
        msg = f"error: expecting an int/None for bits, got {type(bits)} instead"
        raise PyrtlError(msg)

    if not isinstance(axis, int) and axis is not None:
        msg = f"error: expecting an int or None for axis, got {type(axis)} instead"
        raise PyrtlError(msg)

    if bits is None:
        bits = matrix.bits

    if bits <= 0:
        msg = f"error: bits cannot be negative or zero, got {bits} instead"
        raise PyrtlError(msg)

    if axis is None:
        inputs = []
        for i in range(matrix.rows):
            for j in range(matrix.columns):
                inputs.append(matrix[i, j])
        return reduce(sum_2, inputs)

    if axis == 0:
        result = Matrix(1, matrix.columns, signed=matrix.signed, bits=bits)

        for i in range(matrix.columns):
            inputs = []
            for j in range(matrix.rows):
                inputs.append(matrix[j, i])
            result[0, i] = reduce(sum_2, inputs)
        return result

    if axis == 1:
        result = Matrix(1, matrix.rows, signed=matrix.signed, bits=bits)
        for i in range(matrix.rows):
            inputs = []
            for j in range(matrix.columns):
                inputs.append(matrix[i, j])
            result[0, i] = reduce(sum_2, inputs)
        return result

    msg = f"Axis invalid: expected (None, 0, or 1), got {axis}"
    raise PyrtlError(msg)


def min(
    matrix: Matrix | WireVector, axis: int | None = None, bits: int | None = None
) -> Matrix | WireVector:
    """Returns the minimum value in a ``Matrix``.

    This performs a reduction, taking the minimum over the specified ``axis``.

    :param matrix: The matrix to take the mimimum of. If it is a :class:`.WireVector`,
        it will return itself.
    :param axis: The axis to perform the minimum on. ``None`` refers to min of all
        elements. ``0`` is min of columns. ``1`` is min of rows. Defaults to ``None``.
    :param bits: The bits per element of the min. Defaults to ``matrix.bits``.

    :return: A :class:`.WireVector` or ``Matrix`` representing the min value.
    """

    def min_2(first, second):
        return select(first < second, first, second)

    if isinstance(matrix, WireVector):
        return matrix

    if not isinstance(matrix, Matrix):
        msg = (
            f"error: expecting a Matrix or WireVector for matrix, got {type(matrix)} "
            "instead"
        )
        raise PyrtlError(msg)

    if not isinstance(bits, int) and bits is not None:
        msg = f"error: expecting an int/None for bits, got {type(bits)} instead"
        raise PyrtlError(msg)

    if not isinstance(axis, int) and axis is not None:
        msg = f"error: expecting an int or None for axis, got {type(axis)} instead"
        raise PyrtlError(msg)

    if bits is None:
        bits = matrix.bits

    if bits <= 0:
        msg = f"error: bits cannot be negative or zero, got {bits} instead"
        raise PyrtlError(msg)

    if axis is None:
        inputs = []
        for i in range(matrix.rows):
            for j in range(matrix.columns):
                inputs.append(matrix[i, j])
        return reduce(min_2, inputs)

    if axis == 0:
        result = Matrix(1, matrix.columns, signed=matrix.signed, bits=bits)

        for i in range(matrix.columns):
            inputs = []
            for j in range(matrix.rows):
                inputs.append(matrix[j, i])
            result[0, i] = reduce(min_2, inputs)
        return result

    if axis == 1:
        result = Matrix(1, matrix.rows, signed=matrix.signed, bits=bits)
        for i in range(matrix.rows):
            inputs = []
            for j in range(matrix.columns):
                inputs.append(matrix[i, j])
            result[0, i] = reduce(min_2, inputs)
        return result

    msg = f"Axis invalid: expected (None, 0, or 1), got {axis}"
    raise PyrtlError(msg)


def max(
    matrix: Matrix | WireVector, axis: int | None = None, bits: int | None = None
) -> Matrix | WireVector:
    """Returns the maximum value in a ``Matrix``.

    This performs a reduction, taking the maximum over the specified ``axis``.

    :param matrix: The matrix to take the mimimum of. If it is a :class:`.WireVector`,
        it will return itself.
    :param axis: The axis to perform the maximum on. ``None`` refers to max of all
        elements. ``0`` is max of columns. ``1`` is max of rows. Defaults to ``None``.
    :param bits: The bits per element of the max. Defaults to ``matrix.bits``.

    :return: A :class:`.WireVector` or ``Matrix`` representing the max value.
    """

    def max_2(first, second):
        return select(first > second, first, second)

    if isinstance(matrix, WireVector):
        return matrix

    if not isinstance(matrix, Matrix):
        msg = (
            f"error: expecting a Matrix or WireVector for matrix, got {type(matrix)} "
            "instead"
        )
        raise PyrtlError(msg)

    if not isinstance(bits, int) and bits is not None:
        msg = f"error: expecting an int/None for bits, got {type(bits)} instead"
        raise PyrtlError(msg)

    if not isinstance(axis, int) and axis is not None:
        msg = f"error: expecting an int or None for axis, got {type(axis)} instead"
        raise PyrtlError(msg)

    if bits is None:
        bits = matrix.bits

    if bits <= 0:
        msg = f"error: bits cannot be negative or zero, got {bits} instead"
        raise PyrtlError(msg)

    if axis is None:
        inputs = []
        for i in range(matrix.rows):
            for j in range(matrix.columns):
                inputs.append(matrix[i, j])
        return reduce(max_2, inputs)

    if axis == 0:
        result = Matrix(1, matrix.columns, signed=matrix.signed, bits=bits)

        for i in range(matrix.columns):
            inputs = []
            for j in range(matrix.rows):
                inputs.append(matrix[j, i])
            result[0, i] = reduce(max_2, inputs)
        return result

    if axis == 1:
        result = Matrix(1, matrix.rows, signed=matrix.signed, bits=bits)
        for i in range(matrix.rows):
            inputs = []
            for j in range(matrix.columns):
                inputs.append(matrix[i, j])
            result[0, i] = reduce(max_2, inputs)
        return result

    msg = f"Axis invalid: expected (None, 0, or 1), got {axis}"
    raise PyrtlError(msg)


def argmax(
    matrix: Matrix | WireVector, axis: int | None = None, bits: int | None = None
) -> Matrix | WireVector:
    """Returns the index of the max value of the ``Matrix``.

    .. note::

        If there are two indices with the same max value, this function picks the first
        instance.

    :param matrix: The ``Matrix`` to perform argmax operation on. If it is a
        :class:`.WireVector`, it will return itself.
    :param axis: The axis to perform the operation on. ``None`` refers to argmax of all
        items. ``0`` is argmax of the columns. ``1`` is argmax of rows. Defaults to
        ``None``.
    :param bits: The bits per element of the argmax. Defaults to ``matrix.bits``.

    :return: A :class:`.WireVector` or ``Matrix`` representing the argmax value.
    """
    if isinstance(matrix, WireVector):
        return Const(0)

    if not isinstance(matrix, Matrix):
        msg = (
            f"error: expecting a Matrix or WireVector for matrix, got {type(matrix)} "
            "instead"
        )
        raise PyrtlError(msg)

    if not isinstance(bits, int) and bits is not None:
        msg = f"error: expecting an int/None for bits, got {type(bits)} instead"
        raise PyrtlError(msg)

    if not isinstance(axis, int) and axis is not None:
        msg = f"error: expecting an int or None for axis, got {type(axis)} instead"
        raise PyrtlError(msg)

    if bits is None:
        bits = matrix.bits

    if bits <= 0:
        msg = f"error: bits cannot be negative or zero, got {bits} instead"
        raise PyrtlError(msg)

    max_number = max(matrix, axis=axis, bits=bits)
    if axis is None:
        index = Const(0)
        arg = matrix.rows * matrix.columns - 1
        for i in reversed(range(matrix.rows)):
            for j in reversed(range(matrix.columns)):
                index = select(max_number == matrix[i, j], Const(arg), index)
                arg -= 1
        return index
    if axis == 0:
        result = Matrix(1, matrix.columns, signed=matrix.signed, bits=bits)
        for i in range(matrix.columns):
            local_max = max_number[0, i]
            index = Const(0)
            arg = matrix.rows - 1
            for j in reversed(range(matrix.rows)):
                index = select(local_max == matrix[j, i], Const(arg), index)
                arg -= 1
            result[0, i] = index
        return result
    if axis == 1:
        result = Matrix(1, matrix.rows, signed=matrix.signed, bits=bits)
        for i in range(matrix.rows):
            local_max = max_number[0, i]
            index = Const(0)
            arg = matrix.columns - 1
            for j in reversed(range(matrix.columns)):
                index = select(local_max == matrix[i, j], Const(arg), index)
                arg -= 1
            result[0, i] = index
        return result
    return None


def dot(first: Matrix, second: Matrix) -> Matrix:
    """Performs the dot product on two matrices.

    Specifically, the dot product on two matrices is:

    1. If either ``first`` or ``second`` are :class:`WireVectors<.WireVector>`, or have
       both rows and columns equal to 1, ``dot`` is equivalent to :meth:`Matrix.__mul__`

    2. If ``first`` and ``second`` are both arrays (have rows or columns equal to 1),
       ``dot`` is the inner product of the vectors.

    3. Otherwise ``dot`` is :meth:`Matrix.__matmul__` between ``first`` and ``second``.

    .. note::

        Row vectors and column vectors are both treated as arrays.

    :param first: The first matrix.
    :param second: The second matrix.

    :return: A ``Matrix`` that contains the dot product of ``first`` and ``second``.
    """
    if not isinstance(first, (WireVector, Matrix)):
        msg = f"error: expecting a Matrix, got {type(first)} instead"
        raise PyrtlError(msg)
    if not isinstance(second, (WireVector, Matrix)):
        msg = f"error: expecting a Matrix/WireVector, got {type(second)} instead"
        raise PyrtlError(msg)

    # First case when it is multiply
    if isinstance(first, WireVector):
        if isinstance(second, WireVector):
            return first * second
        return second[:, :] * first
    if isinstance(second, WireVector):
        return first[:, :] * second
    if (first.rows == 1 and first.columns == 1) or (
        second.rows == 1 and second.columns == 1
    ):
        return first[:, :] * second[:, :]

    # Second case when it is Inner Product
    if first.rows == 1:
        if second.rows == 1:
            return sum(first * second)
        if second.columns == 1:
            return sum(first * second.transpose())
    elif first.columns == 1:
        if second.rows == 1:
            return sum(first * second.transpose())
        if second.columns == 1:
            return sum(first * second)

    # Third case when it is Matrix Multiply
    return first.__matmul__(second)


def hstack(*matrices: Matrix) -> Matrix:
    """Stack ``matrices`` in sequence horizontally (column-wise).

    All the ``matrices`` must have the same number of rows and the same ``signed``
    value.

    For example::

        m1 = Matrix(rows=2, columns=3, bits=5,
                    value=[[1, 2, 3],
                           [4, 5, 6]])
        m2 = Matrix(rows=2, columns=1, bits=10,
                    value=[[17],
                           [23]]])
        m3 = hstack(m1, m2)

    ``m3`` will look like::

        [[1, 2, 3, 17],
         [4, 5, 6, 23]]

    And ``m3.bits`` will be ``10``.

    :param matrices: Matrices to concatenate together horizontally.

    :return: A new ``Matrix``, with the same number of rows as the original, and columns
             equal to the sum of the columns of ``matrices``. The new ``Matrix``'s
             bitwidth is the max of the bitwidths of all ``matrices``.
    """
    if len(matrices) == 0:
        msg = "Must supply at least one matrix to hstack()"
        raise PyrtlError(msg)

    if any(not isinstance(matrix, Matrix) for matrix in matrices):
        msg = "All arguments to hstack must be matrices."
        raise PyrtlError(msg)

    if len(matrices) == 1:
        return matrices[0].copy()

    new_rows = matrices[0].rows
    if any(m.rows != new_rows for m in matrices):
        msg = "All matrices being hstacked together must have the same number of rows"
        raise PyrtlError(msg)

    new_signed = matrices[0].signed
    if any(m.signed != new_signed for m in matrices):
        msg = "All matrices being hstacked together must have the same signedness"
        raise PyrtlError(msg)

    new_cols = builtins.sum(m.columns for m in matrices)
    new_bits = builtins.max(m.bits for m in matrices)
    new_max_bits = builtins.max(m.max_bits for m in matrices)
    new = Matrix(new_rows, new_cols, new_bits, max_bits=new_max_bits)

    new_c = 0
    for matrix in matrices:
        for c in range(matrix.columns):
            for r in range(matrix.rows):
                new[r, new_c] = matrix[r, c]
            new_c += 1

    return new


def vstack(*matrices: Matrix) -> Matrix:
    """Stack matrices in sequence vertically (row-wise).

    All the ``matrices`` must have the same number of columns and the same ``signed``
    value.

    For example::

        m1 = Matrix(rows=2, columns=3, bits=5,
                    value=[[1, 2, 3],
                           [4, 5, 6]])
        m2 = Matrix(rows=1, columns=3, bits=10,
                    value=[[7, 8, 9]])
        m3 = vstack(m1, m2)

    ``m3`` will look like::

        [[1, 2, 3],
         [4, 5, 6],
         [7, 8, 9]]

    And ``m3.bits`` will be ``10``.

    :param matrices: Matrices to concatenate together vertically

    :return: A new ``Matrix``, with the same number of columns as the original, and rows
             equal to the sum of the rows of ``matricies``. The new ``Matrix``'s
             bitwidth is the max of the bitwidths of all ``matrices``.
    """
    if len(matrices) == 0:
        msg = "Must supply at least one matrix to vstack()"
        raise PyrtlError(msg)

    if any(not isinstance(matrix, Matrix) for matrix in matrices):
        msg = "All arguments to vstack must be matrices."
        raise PyrtlError(msg)

    if len(matrices) == 1:
        return matrices[0].copy()

    new_cols = matrices[0].columns
    if any(m.columns != new_cols for m in matrices):
        msg = (
            "All matrices being vstacked together must have the same number of columns"
        )
        raise PyrtlError(msg)

    new_signed = matrices[0].signed
    if any(m.signed != new_signed for m in matrices):
        msg = "All matrices being vstacked together must have the same signedness"
        raise PyrtlError(msg)

    new_rows = builtins.sum(m.rows for m in matrices)
    new_bits = builtins.max(m.bits for m in matrices)
    new_max_bits = builtins.max(m.max_bits for m in matrices)
    new = Matrix(new_rows, new_cols, new_bits, max_bits=new_max_bits)

    new_r = 0
    for matrix in matrices:
        for r in range(matrix.rows):
            for c in range(matrix.columns):
                new[new_r, c] = matrix[r, c]
            new_r += 1

    return new


def concatenate(matrices: Matrix, axis: int = 0) -> Matrix:
    """Join a sequence of ``matrices`` along an existing ``axis``.

    This function is just a wrapper around :func:`hstack` and :func:`vstack`.

    :param matrices: Matrices to concatenate together.
    :param axis: Axis along which to concatenate. ``0`` is horizontally, ``1`` is
        vertically. Defaults to ``0``.

    :return: A new ``Matrix`` composed of the given matrices concatenated together.
    """
    if axis == 0:
        return hstack(*matrices)
    if axis == 1:
        return vstack(*matrices)
    msg = "Only allowable axes are 0 or 1"
    raise PyrtlError(msg)


def matrix_wv_to_list(
    matrix_wv: WireVector, rows: int, columns: int, bits: int
) -> list[list[int]]:
    """Convert a :class:`.WireVector` representing a :class:`Matrix` into a Python list
    of lists.

    During :class:`.Simulation`, this is useful when printing the value of an
    :meth:`inspected<.Simulation.inspect>` wire that represents a :class:`Matrix`.

    Example::

        m = Matrix.Matrix(rows=2, columns=3, bits=4,
                          values=[[1, 2, 3],
                                  [4, 5, 6]])

        output = Output(name="output")
        output <<= m.to_wirevector()

        sim = Simulation()
        sim.step()

        raw_matrix = Matrix.matrix_wv_to_list(
            sim.inspect("output"), m.rows, m.columns, m.bits)
        print(raw_matrix)

        # Produces:
        # [[1, 2, 3], [4, 5, 6]]

    :param matrix_wv: Result of calling :meth:`Matrix.to_wirevector`.
    :param rows: Number of rows in the matrix.
    :param columns: Number of columns in the matrix.
    :param bits: Number of bits for each element in the matrix.

    :return: A Python list of lists.
    """
    value = bin(matrix_wv)[2:].zfill(rows * columns * bits)

    result = [[0 for _ in range(columns)] for _ in range(rows)]

    bit_pointer = 0
    for i in range(rows):
        for j in range(columns):
            int_value = int(value[bit_pointer : bit_pointer + bits], 2)
            result[i][j] = int_value
            bit_pointer += bits
    return result


def list_to_int(matrix: list[list[int]], n_bits: int) -> int:
    """Convert a Python matrix (a :class:`list` of :class:`lists<list>`) into an
    :class:`int`.

    Integers that are signed will automatically be converted to their two's complement
    form.

    This function is helpful for turning a pure Python list of lists into a very large
    integer suitable for creating a :class:`.Const` that can be used as
    :meth:`Matrix.__init__`'s ``value`` argument, or for passing into a
    :meth:`.Simulation.step`'s ``provided_inputs`` for an :class:`.Input` wire.

    For example, calling ``list_to_int([3, 5], [7, 9], n_bits=4)`` produces ``13689``,
    which in binary looks like::

        0011 0101 0111 1001

    Note how the elements of the list of lists were added, 4 bits at a time, in row
    order, such that the element at row 0, column 0 is in the most significant 4 bits,
    and the element at row 1, column 1 is in the least significant 4 bits.

    Here's an example of using it in :class:`.Simulation`::

        a_vals = [[0, 1], [2, 3]]
        b_vals = [[2, 4, 6], [8, 10, 12]]

        a_in = pyrtl.Input(name="a_in", bitwidth=2 * 2 * 4)
        b_in = pyrtl.Input(name="b_in", bitwidth=2 * 3 * 4)
        a = Matrix.Matrix(rows=2, columns=2, bits=4, value=a_in)
        b = Matrix.Matrix(rows=2, columns=3, bits=4, value=b_in)
        ...

        sim = pyrtl.Simulation()
        sim.step({
            'a_in': Matrix.list_to_int(a_vals, n_bits=a.bits)
            'b_in': Matrix.list_to_int(b_vals, n_bits=b.bits)
        })

    :param matrix: A :class:`list` of :class:`lists<list>` of :class:`ints<int>`
        representing the data in a :class:`Matrix`.
    :param n_bits: The number of bits used to represent each element. If an element
        doesn't fit in ``n_bits``, its most significant bits will be truncated.

    :return: An :class:`int` with bitwidth ``N * n_bits``, containing the elements of
             ``matrix``, where ``N`` is the number of elements in ``matrix``.
    """
    if n_bits <= 0:
        msg = f"Number of bits per element must be positive, instead got {n_bits}"
        raise PyrtlError(msg)

    result = 0

    for i in range(len(matrix)):
        for j in range(len(matrix[0])):
            val = formatted_str_to_val(str(matrix[i][j]), "s" + str(n_bits))
            result = (result << n_bits) | val

    return result



import doctest
import unittest

import pyrtl
from pyrtl.rtllib import muxes
from pyrtl.rtllib import testingutils as utils

gen_in = utils.an_input_and_vals


class TestDocTests(unittest.TestCase):
    """Test documentation examples."""

    def test_doctests(self):
        failures, tests = doctest.testmod(m=pyrtl.rtllib.muxes)
        self.assertGreater(tests, 0)
        self.assertEqual(failures, 0)


class TestPrioritizedMuxTrivial(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        pyrtl.reset_working_block()

    def test_empty(self):
        with self.assertRaises(pyrtl.PyrtlError):
            _ = muxes.prioritized_mux([], [])

    def test_different_sel_and_val_lengths(self):
        a = pyrtl.WireVector(1)
        with self.assertRaises(pyrtl.PyrtlError):
            _ = muxes.prioritized_mux([a], [a, a])

    def test_invalid_select_width(self):
        a = pyrtl.WireVector(2)
        b = pyrtl.WireVector(2)
        c = pyrtl.WireVector(10)
        with self.assertRaises(pyrtl.PyrtlError):
            _ = muxes.prioritized_mux([a, b], [c, c])

    def test_one_wire(self):
        a = pyrtl.WireVector(1)
        b = pyrtl.WireVector(10)
        x = muxes.prioritized_mux([a], [b])
        self.assertIs(b, x)


def pri_mux_actual(sels, vals):
    # python version of the pri mux hardware
    assert len(sels) == len(vals)
    for index, s in enumerate(sels):
        if s:
            return vals[index]
    return vals[-1]


class TestPrioritizedMuxSim(unittest.TestCase):
    def setUp(self):
        pyrtl.reset_working_block()

    def test_select_with_2_wires(self):
        val_width = 5
        sels, sel_vals = utils.make_inputs_and_values(2, exact_bitwidth=1)
        mux_ins, vals = utils.make_inputs_and_values(2, exact_bitwidth=val_width)

        out = pyrtl.Output(val_width, "out")
        out <<= muxes.prioritized_mux(sels, mux_ins)
        actual = utils.sim_and_ret_out(out, sels + mux_ins, sel_vals + vals)
        expected = [
            pri_mux_actual(sel, val) for sel, val in zip(zip(*sel_vals), zip(*vals))
        ]
        self.assertEqual(actual, expected)

    def test_select_with_5_wires(self):
        val_width = 5
        sels, sel_vals = utils.make_inputs_and_values(5, exact_bitwidth=1, test_vals=50)
        mux_ins, vals = utils.make_inputs_and_values(
            5, exact_bitwidth=val_width, test_vals=50
        )

        out = pyrtl.Output(val_width, "out")
        out <<= muxes.prioritized_mux(sels, mux_ins)
        actual = utils.sim_and_ret_out(out, sels + mux_ins, sel_vals + vals)
        expected = [
            pri_mux_actual(sel, val) for sel, val in zip(zip(*sel_vals), zip(*vals))
        ]
        self.assertEqual(actual, expected)


class TestIsEquivalent(unittest.TestCase):
    def test_equivalent_const(self):
        a = pyrtl.Const(1)
        b = pyrtl.Const(1)
        c = pyrtl.Const(1, 2)
        d = pyrtl.Const(3)
        self.assertTrue(muxes._is_equivalent(a, a))
        self.assertTrue(muxes._is_equivalent(a, b))
        self.assertFalse(muxes._is_equivalent(a, c))
        self.assertFalse(muxes._is_equivalent(a, d))

    def test_equivalent(self):
        a = pyrtl.WireVector(2)
        b = pyrtl.Const(2, 2)
        c = pyrtl.Output()

        self.assertTrue(muxes._is_equivalent(a, a))
        self.assertTrue(muxes._is_equivalent(c, c))
        self.assertFalse(muxes._is_equivalent(a, b))
        self.assertFalse(muxes._is_equivalent(a, c))


class TestSmartMuxTrivial(unittest.TestCase):
    def setUp(self):
        pyrtl.reset_working_block()

    def test_one_value(self):
        sel = pyrtl.WireVector(3)
        a = pyrtl.WireVector(1)
        self.assertIs(muxes.sparse_mux(sel, {6: a}), a)

    def test_dup_value(self):
        sel = pyrtl.WireVector(3)
        a = pyrtl.WireVector(1)
        self.assertIs(muxes.sparse_mux(sel, {6: a, 2: a}), a)

    def test_dup_consts1(self):
        sel = pyrtl.WireVector(1)
        c1 = pyrtl.Const(4)
        c2 = pyrtl.Const(4)
        res = muxes.sparse_mux(sel, {0: c1, 1: c2})
        self.assertIsInstance(res, pyrtl.Const)
        self.assertEqual(res.val, 4)

    def test_dup_consts2(self):
        sel = pyrtl.WireVector(3)
        c1 = pyrtl.Const(4)
        c2 = pyrtl.Const(4)
        res = muxes.sparse_mux(sel, {6: c1, 2: c2})
        self.assertIsInstance(res, pyrtl.Const)
        self.assertEqual(res.val, 4)

    def test_no_dup_2(self):
        sel = pyrtl.WireVector(3)
        c1 = pyrtl.Const(4)
        c2 = pyrtl.Const(6)
        res = muxes.sparse_mux(sel, {6: c1, 2: c2})
        self.assertNotIsInstance(res, pyrtl.Const)

    def test_no_dup(self):
        sel = pyrtl.WireVector(3)
        a = pyrtl.WireVector(3)
        b = pyrtl.WireVector(3)
        res = muxes.sparse_mux(sel, {6: a, 2: b})
        self.assertIsNot(res, a)
        self.assertIsNot(res, b)


class TestSmartMux(unittest.TestCase):
    def setUp(self):
        pyrtl.reset_working_block()

    def test_two_vals(self):
        sel, sel_vals = gen_in(1)
        a1, a1_vals = gen_in(3)
        a2, a2_vals = gen_in(3)
        res = pyrtl.Output(name="output")

        res <<= muxes.sparse_mux(sel, {0: a1, 1: a2})
        in_vals = [sel_vals, a1_vals, a2_vals]
        out_res = utils.sim_and_ret_out(res, [sel, a1, a2], in_vals)

        expected_out = [e2 if sel else e1 for sel, e1, e2 in zip(*in_vals)]
        self.assertEqual(out_res, expected_out)

    def test_two_vals_big(self):
        sel = pyrtl.Input(3)
        a1, a1_vals = gen_in(3)
        a2, a2_vals = gen_in(3)
        res = pyrtl.Output(name="output")

        sel_vals = [utils.uniform_dist(1) for i in range(20)]
        real_sel = [6 if s else 2 for s in sel_vals]
        res <<= muxes.sparse_mux(sel, {2: a1, 6: a2})
        out_res = utils.sim_and_ret_out(
            res, [sel, a1, a2], [real_sel, a1_vals, a2_vals]
        )

        expected_out = [
            e2 if sel else e1 for sel, e1, e2 in zip(sel_vals, a1_vals, a2_vals)
        ]
        self.assertEqual(out_res, expected_out)

    def test_multiple_bitwidths(self):
        sel = pyrtl.Input(3)
        a1, a1_vals = gen_in(3)
        a2, a2_vals = gen_in(8)
        a3, a3_vals = gen_in(5)

        m = muxes.sparse_mux(sel, {2: a1, 3: a2, 6: a3})
        self.assertEqual(len(m), 8)  # the biggest one

    def test_two_big_close(self):
        sel = pyrtl.Input(3)
        a1, a1_vals = gen_in(3)
        a2, a2_vals = gen_in(3)
        res = pyrtl.Output(name="output")

        sel_vals = [utils.uniform_dist(1) for i in range(20)]
        real_sel = [6 if s else 5 for s in sel_vals]
        res <<= muxes.sparse_mux(sel, {5: a1, 6: a2})
        out_res = utils.sim_and_ret_out(
            res, [sel, a1, a2], [real_sel, a1_vals, a2_vals]
        )

        expected_out = [
            e2 if sel else e1 for sel, e1, e2 in zip(sel_vals, a1_vals, a2_vals)
        ]
        self.assertEqual(out_res, expected_out)


class TestSmartMuxDefault(unittest.TestCase):
    def setUp(self):
        pyrtl.reset_working_block()

    def test_default(self):
        sel, sel_vals = gen_in(3)
        a1, a1_vals = gen_in(3)
        a2, a2_vals = gen_in(3)
        default, default_vals = gen_in(3)
        res = pyrtl.Output(name="output")

        res <<= muxes.sparse_mux(sel, {5: a1, 6: a2, muxes.SparseDefault: default})
        out_res = utils.sim_and_ret_out(
            res, [sel, a1, a2, default], [sel_vals, a1_vals, a2_vals, default_vals]
        )

        expected_out = [
            e2 if sel == 6 else e1 if sel == 5 else d
            for sel, e1, e2, d in zip(sel_vals, a1_vals, a2_vals, default_vals)
        ]
        self.assertEqual(out_res, expected_out)


class TestMultiSelector(unittest.TestCase):
    def setUp(self):
        pyrtl.reset_working_block()

    def test_value_already_set(self):
        sel = pyrtl.Input(1)
        wire = pyrtl.WireVector(8)
        i1_out = pyrtl.Output(name="i1_out")
        i2_out = pyrtl.Output(name="i2_out")
        with muxes.MultiSelector(sel, i1_out, i2_out) as mul_sel:
            mul_sel.option(0, wire, wire)
            with self.assertRaises(pyrtl.PyrtlError):
                mul_sel.option(0, wire, wire)

    def test_incorrect_number_of_wires(self):
        sel = pyrtl.Input(1)
        wire = pyrtl.WireVector(8)
        i1_out = pyrtl.Output(name="i1_out")
        i2_out = pyrtl.Output(name="i2_out")
        with muxes.MultiSelector(sel, i1_out, i2_out) as mul_sel:
            mul_sel.option(0, wire, wire)
            with self.assertRaises(pyrtl.PyrtlError):
                mul_sel.option(1, wire, wire, wire)

    def test_incorrect_number_of_wires_2(self):
        sel = pyrtl.Input(1)
        wire = pyrtl.WireVector(8)
        i1_out = pyrtl.Output(name="i1_out")
        i2_out = pyrtl.Output(name="i2_out")
        i3_out = pyrtl.Output(name="i3_out")
        mul_sel = muxes.MultiSelector(sel, i1_out, i2_out, i3_out)
        with self.assertRaises(pyrtl.PyrtlError):
            mul_sel.option(0, wire, wire)


class TestMultiSelectorSim(unittest.TestCase):
    def setUp(self):
        pyrtl.reset_working_block()

    def test_really_simple(self):
        sel, sel_vals = gen_in(1)

        i1_0, i1_0_vals = gen_in(8)
        i2_0, i2_0_vals = gen_in(8)
        i1_1, i1_1_vals = gen_in(8)
        i2_1, i2_1_vals = gen_in(8)

        i1_out = pyrtl.Output(name="i1_out")
        i2_out = pyrtl.Output(name="i2_out")

        with muxes.MultiSelector(sel, i1_out, i2_out) as mul_sel:
            mul_sel.option(0, i1_0, i2_0)
            mul_sel.option(1, i1_1, i2_1)

        actual_outputs = utils.sim_and_ret_outws(
            [sel, i1_0, i1_1, i2_0, i2_1],
            [sel_vals, i1_0_vals, i1_1_vals, i2_0_vals, i2_1_vals],
        )
        expected_i1_out = [
            v1 if s else v0 for s, v0, v1 in zip(sel_vals, i1_0_vals, i1_1_vals)
        ]
        expected_i2_out = [
            v1 if s else v0 for s, v0, v1 in zip(sel_vals, i2_0_vals, i2_1_vals)
        ]

        self.assertEqual(actual_outputs[i1_out.name], expected_i1_out)
        self.assertEqual(actual_outputs[i2_out.name], expected_i2_out)

    def test_simple(self):
        sel, sel_vals = gen_in(2)

        x1s, x1_vals = (list(x) for x in zip(*(gen_in(8) for i in range(4))))
        x2s, x2_vals = (list(x) for x in zip(*(gen_in(8) for i in range(4))))
        x3s, x3_vals = (list(x) for x in zip(*(gen_in(8) for i in range(4))))

        i1_out = pyrtl.Output(name="i1_out")
        i2_out = pyrtl.Output(name="i2_out")
        i3_out = pyrtl.Output(name="i3_out")

        with muxes.MultiSelector(sel, i1_out, i2_out, i3_out) as mu:
            for i in range(4):
                mu.option(i, x1s[i], x2s[i], x3s[i])

        wires = [sel, *x1s, *x2s, *x3s]
        vals = [sel_vals, *x1_vals, *x2_vals, *x3_vals]
        actual_outputs = utils.sim_and_ret_outws(wires, vals)

        expected_i1_out = [v[s] for s, v in zip(sel_vals, zip(*x1_vals))]
        expected_i2_out = [v[s] for s, v in zip(sel_vals, zip(*x2_vals))]
        expected_i3_out = [v[s] for s, v in zip(sel_vals, zip(*x3_vals))]

        self.assertEqual(actual_outputs[i1_out.name], expected_i1_out)
        self.assertEqual(actual_outputs[i2_out.name], expected_i2_out)
        self.assertEqual(actual_outputs[i3_out.name], expected_i3_out)


class TestDemux(unittest.TestCase):
    def setUp(self):
        pyrtl.reset_working_block()

    def test_simple_demux(self):
        in_w, in_vals = utils.an_input_and_vals(2)
        outs = (pyrtl.Output(name="output_" + str(i)) for i in range(4))
        demux_outs = pyrtl.rtllib.muxes.demux(in_w)
        for out_w, demux_out in zip(outs, demux_outs):
            out_w <<= demux_out
        traces = utils.sim_and_ret_outws((in_w,), (in_vals,))

        for cycle in range(20):
            for i, out_wire in enumerate(outs):
                self.assertEqual(in_vals[i] == i, traces[out_wire][cycle])

    def test_demux_2(self):
        in_w, in_vals = utils.an_input_and_vals(1)
        outs = (pyrtl.Output(name="output_" + str(i)) for i in range(2))
        demux_outs = pyrtl.rtllib.muxes._demux_2(in_w)
        for out_w, demux_out in zip(outs, demux_outs):
            out_w <<= demux_out
        traces = utils.sim_and_ret_outws((in_w,), (in_vals,))

        for cycle in range(20):
            for i, out_wire in enumerate(outs):
                self.assertEqual(in_vals[i] == i, traces[out_wire][cycle])

    def test_large_demux(self):
        in_w, in_vals = utils.an_input_and_vals(5)
        outs = (pyrtl.Output(name="output_" + str(i)) for i in range(32))
        demux_outs = pyrtl.rtllib.muxes.demux(in_w)
        for out_w, demux_out in zip(outs, demux_outs):
            self.assertEqual(len(demux_out), 1)
            out_w <<= demux_out
        traces = utils.sim_and_ret_outws((in_w,), (in_vals,))

        for cycle in range(20):
            for i, out_wire in enumerate(outs):
                self.assertEqual(in_vals[i] == i, traces[out_wire][cycle])


if __name__ == "__main__":
    unittest.main()



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

sys.path.insert(0, os.path.abspath(".."))


# -- Project information -----------------------------------------------------

project = "PyRTL"
copyright = "2025, Timothy Sherwood"
author = "Timothy Sherwood"

# -- General configuration ---------------------------------------------------

master_doc = "index"

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.inheritance_diagram",
    "sphinx.ext.intersphinx",
    "sphinx.ext.viewcode",
    "sphinx_autodoc_typehints",
    "sphinx_copybutton",
]

graphviz_output_format = "svg"

# Omit redundant method names in right sidebar (step() instead of Simulation.step()).
toc_object_entries_show_parents = "hide"

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# Enable links to Python standard library classes (str, list, dict, etc).
intersphinx_mapping = {"python": ("https://docs.python.org/3", None)}

# sphinx_copybutton excludes line numbers, prompts, and outputs.
copybutton_exclude = ".linenos, .gp, .go"

# sphinx-autodoc-typehints configuration: Always display Unions with vertical bars,
# show default values, and don't document :rtype: None.
always_use_bars_union = True
typehints_defaults = "comma"
typehints_document_rtype_none = False

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "furo"
html_theme_options = {
    "sidebar_hide_name": True,
    # For view/edit this page buttons.
    "source_repository": "https://github.com/UCSBarchlab/pyrtl",
    "source_branch": "development",
    "source_directory": "docs/",
    # Add a GitHub repository link to the footer.
    "footer_icons": [
        {
            "name": "GitHub",
            "url": "https://github.com/UCSBarchlab/pyrtl",
            "html": """
                <svg stroke="currentColor" fill="currentColor" stroke-width="0" viewBox="0 0 16 16">
                    <path fill-rule="evenodd" d="M8 0C3.58 0 0 3.58 0 8c0 3.54 2.29 6.53 5.47 7.59.4.07.55-.17.55-.38 0-.19-.01-.82-.01-1.49-2.01.37-2.53-.49-2.69-.94-.09-.23-.48-.94-.82-1.13-.28-.15-.68-.52-.01-.53.63-.01 1.08.58 1.23.82.72 1.21 1.87.87 2.33.66.07-.52.28-.87.51-1.07-1.78-.2-3.64-.89-3.64-3.95 0-.87.31-1.59.82-2.15-.08-.2-.36-1.02.08-2.12 0 0 .67-.21 2.2.82.64-.18 1.32-.27 2-.27.68 0 1.36.09 2 .27 1.53-1.04 2.2-.82 2.2-.82.44 1.1.16 1.92.08 2.12.51.56.82 1.27.82 2.15 0 3.07-1.87 3.75-3.65 3.95.29.25.54.73.54 1.48 0 1.07-.01 1.93-.01 2.2 0 .21.15.46.55.38A8.013 8.013 0 0 0 16 8c0-4.42-3.58-8-8-8z"></path>
                </svg>
            """,  # noqa: E501
            "class": "",
        },
    ],
}
html_logo = "brand/pyrtl_logo.png"

# Force a light blue background color for inheritance-diagrams. The default is
# transparent, which does not work well with Furo's dark mode.
inheritance_graph_attrs = {
    "bgcolor": "aliceblue",
}



"""
Basic multiplexers are defined in PyRTL's core library, see:

- :func:`.select` for a multiplexer that selects between two options.

- :func:`.mux` for a multiplexer that selects between an arbitrary number of options.

- :ref:`conditional_assignment` for a more readable alternative to nested
  :func:`selects<.select>` and :func:`muxes<.mux>`.

The functions below provide more complex alternatives.
"""

import numbers

import pyrtl
from pyrtl import WireVector


def prioritized_mux(selects: list[WireVector], vals: list[WireVector]) -> WireVector:
    """Returns the value in the first wire for which its ``select`` bit is ``1``

    If none of the ``selects`` are ``1``, the last ``val`` is returned.

    .. doctest only::

        >>> import pyrtl
        >>> pyrtl.reset_working_block()

    Example::

        >>> selects = [pyrtl.Input(name=f"select{i}", bitwidth=1)
        ...            for i in range(3)]
        >>> vals = [pyrtl.Const(n) for n in range(2, 5)]
        >>> output = pyrtl.Output(name="output")

        >>> output <<= pyrtl.rtllib.muxes.prioritized_mux(selects, vals)

        >>> sim = pyrtl.Simulation()
        >>> sim.step(provided_inputs={"select0": 1, "select1": 0, "select2": 0})
        >>> sim.inspect("output")
        2

        >>> sim.step(provided_inputs={"select0": 0, "select1": 1, "select2": 0})
        >>> sim.inspect("output")
        3

        >>> sim.step(provided_inputs={"select0": 0, "select1": 0, "select2": 0})
        >>> sim.inspect("output")
        4

    :param selects: A list of :class:`WireVectors<.WireVector>` signaling whether a wire
        should be chosen.
    :param vals: Values to return when the corresponding ``select`` value is ``1``.

    :return: The selected value.
    """
    if len(selects) != len(vals):
        msg = "Number of select and val signals must match"
        raise pyrtl.PyrtlError(msg)
    if len(vals) == 0:
        msg = "Must have a signal to mux"
        raise pyrtl.PyrtlError(msg)
    if len(vals) == 1:
        return vals[0]
    half = len(vals) // 2
    return pyrtl.select(
        pyrtl.rtl_any(*selects[:half]),
        truecase=prioritized_mux(selects[:half], vals[:half]),
        falsecase=prioritized_mux(selects[half:], vals[half:]),
    )


def _is_equivalent(w1, w2):
    if isinstance(w1, pyrtl.Const) & isinstance(w2, pyrtl.Const):
        return (w1.val == w2.val) & (w1.bitwidth == w2.bitwidth)
    return w1 is w2


SparseDefault = "default"
"""
A special key for :func:`sparse_mux`'s ``vals`` :class:`dict` that specifies the mux's
default value.
"""


def sparse_mux(sel: WireVector, vals: dict[int, WireVector]) -> WireVector:
    """Mux that avoids instantiating unnecessary ``mux_2s`` when possible.

    ``sparse_mux`` supports not having a full specification. Indices that are not
    specified are treated as don't-cares.

    .. doctest only::

        >>> import pyrtl
        >>> pyrtl.reset_working_block()

    Example::

        >>> select = pyrtl.Input(name="select", bitwidth=3)
        >>> vals = {2: pyrtl.Const(3),
        ...         4: pyrtl.Const(5),
        ...         pyrtl.rtllib.muxes.SparseDefault: pyrtl.Const(7)}
        >>> output = pyrtl.Output(name="output")

        >>> output <<= pyrtl.rtllib.muxes.sparse_mux(select, vals)

        >>> sim = pyrtl.Simulation()
        >>> sim.step(provided_inputs={"select": 2})
        >>> sim.inspect("output")
        3

        >>> sim.step(provided_inputs={"select": 4})
        >>> sim.inspect("output")
        5

        >>> sim.step(provided_inputs={"select": 3})
        >>> sim.inspect("output")
        7

    :param sel: Select wire, which chooses one of the mux input ``vals`` to output.
    :param vals: :class:`dict` of mux input values. If the special key
        :data:`SparseDefault` exists, it specifies the ``sparse_mux``'s default value.

    :return: The :class:`.WireVector` selected from ``vals`` by ``sel``.
    """
    max_val = 2 ** len(sel) - 1
    if SparseDefault in vals:
        default_val = vals[SparseDefault]
        del vals[SparseDefault]
        for i in range(max_val + 1):
            if i not in vals:
                vals[i] = default_val

    for key in vals:
        if not isinstance(key, numbers.Integral):
            msg = f"value {key} must be either an integer or 'default'"
            raise pyrtl.PyrtlError(msg)
        if key < 0 or key > max_val:
            msg = f"value {key} is out of range of the sel wire"
            raise pyrtl.PyrtlError(msg)

    return _sparse_mux(sel, vals)


def _sparse_mux(sel, vals):
    """Mux that avoids instantiating unnecessary mux_2s when possible.

    This mux supports not having a full specification. indices that are not specified
    are treated as Don't Cares

    :param WireVector sel: Select wire, determines what is selected on a given cycle
    :param {int: WireVector} vals: dictionary to store the values that are

    :return: WireVector that signifies the change
    """
    items = list(vals.values())
    if len(vals) <= 1:
        if len(vals) == 0:
            msg = "Needs at least one parameter for val"
            raise pyrtl.PyrtlError(msg)
        return items[0]

    if len(sel) == 1:
        try:
            false_result = vals[0]
            true_result = vals[1]
        except KeyError as exc:
            msg = (
                "Failed to retrieve values for smartmux. The length of sel might be "
                "wrong"
            )
            raise pyrtl.PyrtlError(msg) from exc
    else:
        half = 2 ** (len(sel) - 1)

        first_dict = {indx: wire for indx, wire in vals.items() if indx < half}
        second_dict = {indx - half: wire for indx, wire in vals.items() if indx >= half}
        if not first_dict:
            return sparse_mux(sel[:-1], second_dict)
        if not second_dict:
            return sparse_mux(sel[:-1], first_dict)

        false_result = sparse_mux(sel[:-1], first_dict)
        true_result = sparse_mux(sel[:-1], second_dict)
    if _is_equivalent(false_result, true_result):
        return true_result
    return pyrtl.select(sel[-1], falsecase=false_result, truecase=true_result)


class MultiSelector:
    """The MultiSelector allows you to specify multiple wire value results for a single
    select wire.

    Useful for processors, finite state machines and other places where the result of
    many wire values are determined by a common wire signal (such as a 'state' wire).

    Example::

        with muxes.MultiSelector(select, res0, res1, res2, ...) as ms:
            ms.option(val1, data0, data1, data2, ...)
            ms.option(val2, data0_2, data1_2, data2_2, ...)

    This means that when the ``select`` wire equals the ``val1`` wire the results will
    have the values in ``data0, data1, data2, ...`` (all ints are converted to wires)

    .. WARNING::

        Use :ref:`conditional_assignment` instead.
    """

    def __init__(self, signal_wire, *dest_wires):
        self._final = False
        self.dest_wires = dest_wires
        self.signal_wire = signal_wire
        self.instructions = []
        self.dest_instrs_info = {dest_w: [] for dest_w in dest_wires}

    def __enter__(self):
        """For compatibility with `with` statements, which is the recommended method of
        using a MultiSelector.
        """
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is None:
            self.finalize()
        else:
            print("The MultiSelector was not finalized due to uncaught exception")

    def _check_finalized(self):
        if self._final:
            msg = "Cannot change InstrConnector, already finalized"
            raise pyrtl.PyrtlError(msg)

    def option(self, select_val, *data_signals):
        self._check_finalized()
        instr, ib = pyrtl.infer_val_and_bitwidth(select_val, self.signal_wire.bitwidth)
        if instr in self.instructions:
            msg = f"instruction {select_val} already exists"
            raise pyrtl.PyrtlError(msg)
        self.instructions.append(instr)
        self._add_signal(data_signals)

    def default(self, *data_signals):
        self._check_finalized()
        self.instructions.append(SparseDefault)
        self._add_signal(data_signals)

    def _add_signal(self, data_signals):
        self._check_finalized()
        if len(data_signals) != len(self.dest_wires):
            msg = (
                "Incorrect number of data_signals for instruction received "
                f"{len(data_signals)}, expected {len(self.dest_wires)}"
            )
            raise pyrtl.PyrtlError(msg)

        for dw, sig in zip(self.dest_wires, data_signals):
            data_signal = pyrtl.as_wires(sig, dw.bitwidth)
            self.dest_instrs_info[dw].append(data_signal)

    def finalize(self):
        """Connects the wires."""
        self._check_finalized()
        self._final = True

        for dest_w, values in self.dest_instrs_info.items():
            mux_vals = dict(zip(self.instructions, values))
            dest_w <<= sparse_mux(self.signal_wire, mux_vals)


def demux(select: WireVector) -> tuple[WireVector, ...]:
    """Demultiplexes a wire of arbitrary bitwidth.

    This effectively converts an unsigned binary value into a one-hot encoded value,
    returning each bit of the one-hot encoded value as a separate :class:`.WireVector`.

    .. doctest only::

        >>> import pyrtl
        >>> pyrtl.reset_working_block()

    Example::

        >>> input = pyrtl.Input(bitwidth=3)

        >>> outputs = pyrtl.rtllib.muxes.demux(input)
        >>> len(outputs)
        8
        >>> len(outputs[0])
        1
        >>> for i, wire in enumerate(outputs):
        ...     wire.name = f"outputs[{i}]"

        >>> sim = pyrtl.Simulation()
        >>> sim.step(provided_inputs={input.name: 5})

        >>> sim.inspect("outputs[4]")
        0
        >>> sim.inspect("outputs[5]")
        1
        >>> sim.inspect("outputs[6]")
        0

    In the example above, ``len(outputs)`` is ``8`` because ``2 ** 3 == 8``, and
    ``outputs[5]`` is ``1`` because the output index ``5`` matches the input value.

    See :func:`.binary_to_one_hot`, which performs a similar operation.

    .. WARNING::

        ``demux`` can create a very large number of ``WireVectors``. Use with
        caution.

    :param select: The value to demultiplex.

    :return: A tuple of 1-bit wires, where each wire indicates if the value of
             ``select`` equals the wire's index in the tuple. The tuple has length ``2
             ** select.bitwidth``.
    """
    if len(select) == 1:
        return _demux_2(select)

    wires = demux(select[:-1])
    sel = select[-1]
    not_select = ~sel
    zero_wires = tuple(not_select & w for w in wires)
    one_wires = tuple(sel & w for w in wires)
    return zero_wires + one_wires


def _demux_2(select):
    assert len(select) == 1
    return ~select, select



import doctest
import unittest

import pyrtl


class TestDocTests(unittest.TestCase):
    """Test documentation examples."""

    def test_doctests(self):
        failures, tests = doctest.testmod(m=pyrtl.gate_graph)
        self.assertGreater(tests, 0)
        self.assertEqual(failures, 0)


class TestGateGraph(unittest.TestCase):
    def setUp(self):
        pyrtl.reset_working_block()

    def test_gate_retrieval(self):
        a = pyrtl.Input(name="a", bitwidth=1)
        b = pyrtl.Input(name="b", bitwidth=1)
        c = pyrtl.Input(name="c", bitwidth=2)
        ab = a + b
        ab.name = "ab"
        abc = ab - c
        abc.name = "abc"

        gate_graph = pyrtl.GateGraph()

        self.assertEqual(
            sorted(gate.name for gate in gate_graph),
            ["a", "ab", "abc", "b", "c"],
        )

        self.assertEqual(
            sorted([gate.name for gate in gate_graph.gates]),
            ["a", "ab", "abc", "b", "c"],
        )

        self.assertEqual(
            sorted([gate.name for gate in gate_graph.sources]), ["a", "b", "c"]
        )

        self.assertEqual(sorted([gate.name for gate in gate_graph.sinks]), ["abc"])

        gate_ab = gate_graph.get_gate("ab")
        self.assertEqual(gate_ab.name, "ab")
        self.assertEqual(gate_ab.op, "+")

        gate_abc = gate_graph.get_gate("abc")
        self.assertEqual(gate_abc.name, "abc")
        self.assertEqual(gate_abc.op, "-")

    def test_get_gate(self):
        _ = pyrtl.Input(name="a", bitwidth=4)

        gate_graph = pyrtl.GateGraph()
        a_gate = gate_graph.get_gate("a")
        self.assertEqual(a_gate.name, "a")

        self.assertEqual(gate_graph.get_gate("q"), None)

    def test_select_gate(self):
        a = pyrtl.Input(name="a", bitwidth=4)
        b = pyrtl.Input(name="b", bitwidth=4)
        s = pyrtl.Input(name="s", bitwidth=1)

        output = pyrtl.select(s, a, b)
        output.name = "output"

        gate_graph = pyrtl.GateGraph()
        select_gate = gate_graph.get_gate("output")
        self.assertEqual(select_gate.op, "x")
        self.assertEqual(select_gate.op_param, None)
        a_gate = gate_graph.get_gate("a")
        b_gate = gate_graph.get_gate("b")
        s_gate = gate_graph.get_gate("s")
        self.assertEqual(select_gate.args, [s_gate, b_gate, a_gate])
        self.assertEqual(select_gate.name, "output")
        self.assertEqual(select_gate.bitwidth, 4)
        self.assertEqual(select_gate.dests, [])
        self.assertEqual(str(select_gate), "output/4 = s/1 ? a/4 : b/4")

    def test_gate_attrs(self):
        a = pyrtl.Input(name="a", bitwidth=4)
        b = pyrtl.Const(name="b", bitwidth=2, val=1)
        bit_slice = a[2:4]
        bit_slice.name = "bit_slice"
        ab = bit_slice + b
        ab.name = "ab"

        output = pyrtl.Output(name="output", bitwidth=3)
        bb = b + b
        bb.name = "bb"
        output <<= bb

        gate_graph = pyrtl.GateGraph()

        a_gate = gate_graph.get_gate("a")
        self.assertEqual(a_gate.op, "I")
        self.assertEqual(a_gate.args, [])

        b_gate = gate_graph.get_gate("b")
        self.assertEqual(b_gate.op, "C")
        self.assertEqual(b_gate.op_param, (1,))

        self.assertEqual(str(b_gate), "b/2 = Const(1)")

        bit_slice_gate = gate_graph.get_gate("bit_slice")
        self.assertEqual(bit_slice_gate.op, "s")

        self.assertEqual(bit_slice_gate.op_param, (2, 3))

        self.assertEqual(bit_slice_gate.args, [a_gate])

        self.assertEqual(bit_slice_gate.name, "bit_slice")
        self.assertEqual(bit_slice_gate.bitwidth, 2)

        self.assertEqual(str(bit_slice_gate), "bit_slice/2 = slice(a/4) [sel=(2, 3)]")

        ab_gate = gate_graph.get_gate("ab")
        self.assertEqual(ab_gate.op, "+")

        self.assertEqual(ab_gate.args, [bit_slice_gate, b_gate])

        self.assertEqual(ab_gate.name, "ab")
        self.assertEqual(ab_gate.bitwidth, 3)
        self.assertFalse(ab_gate.is_output)

        output_gate = gate_graph.get_gate("output")
        self.assertEqual(output_gate.op, "w")
        self.assertTrue(output_gate.is_output)
        self.assertEqual(str(output_gate), "output/3 [Output] = bb/3")

        self.assertEqual(len(output_gate.args), 1)
        output_add_gate = output_gate.args[0]

        self.assertEqual(output_add_gate.args, [b_gate, b_gate])

        self.assertEqual(len(b_gate.dests), 3)
        num_ab_gates = 0
        num_output_add_gates = 0
        for dest_gate in b_gate.dests:
            if dest_gate is ab_gate:
                num_ab_gates += 1
            elif dest_gate is output_add_gate:
                num_output_add_gates += 1
        self.assertEqual(num_ab_gates, 1)
        self.assertEqual(num_output_add_gates, 2)

    def test_register_gate_forward(self):
        counter = pyrtl.Register(name="counter", bitwidth=3)
        one = pyrtl.Const(name="one", bitwidth=3, val=1)
        truncated = (counter + one).truncate(3)
        truncated.name = "truncated"
        counter.next <<= truncated

        gate_graph = pyrtl.GateGraph()

        # Traverse the ``GateGraph`` forward, following ``dests`` references, from
        # ``counter``. We should end up back at ``counter``.
        counter_gate = gate_graph.get_gate("counter")
        self.assertEqual(len(counter_gate.dests), 1)
        self.assertEqual(
            str(counter_gate), "counter/3 = reg(truncated/3) [reset_value=0]"
        )

        plus_gate = counter_gate.dests[0]
        self.assertEqual(plus_gate.op, "+")
        self.assertEqual(len(plus_gate.dests), 1)

        slice_gate = plus_gate.dests[0]
        self.assertEqual(slice_gate.op, "s")
        self.assertEqual(len(slice_gate.dests), 1)

        self.assertEqual(slice_gate.dests[0], counter_gate)

    def test_register_gate_backward(self):
        counter = pyrtl.Register(name="counter", bitwidth=3, reset_value=2)
        one = pyrtl.Const(name="one", bitwidth=3, val=1)
        counter.next <<= counter + one

        gate_graph = pyrtl.GateGraph()

        # Traverse the ``GateGraph`` backward, following ``args`` references, from
        # ``counter``. We should end up back at ``counter``.
        counter_gate = gate_graph.get_gate("counter")
        self.assertEqual(len(counter_gate.args), 1)
        self.assertEqual(counter_gate.op_param, (2,))

        # Implicit truncation from 4-bit sum to 3-bit register input.
        slice_gate = counter_gate.args[0]
        self.assertEqual(slice_gate.op, "s")
        self.assertEqual(len(slice_gate.args), 1)

        plus_gate = slice_gate.args[0]
        self.assertEqual(plus_gate.op, "+")
        self.assertEqual(len(plus_gate.args), 2)

        self.assertEqual(plus_gate.args[0], counter_gate)

    def test_register_self_loop(self):
        """Test a register that sets its next value directly from itself.

        This is an unusual case that creates a self-loop in the ``GateGraph``.
        """
        r = pyrtl.Register(name="r", bitwidth=1)
        r.next <<= r

        gate_graph = pyrtl.GateGraph()
        r_gate = gate_graph.get_gate("r")
        self.assertEqual(r_gate.args[0], r_gate)
        self.assertEqual(r_gate.dests[0], r_gate)
        self.assertEqual(str(r_gate), "r/1 = reg(r/1) [reset_value=0]")

    def test_memblock(self):
        mem = pyrtl.MemBlock(name="mem", bitwidth=8, addrwidth=2)

        write_addr = pyrtl.Input(name="write_addr", bitwidth=2)
        write_data = pyrtl.Input(name="write_data", bitwidth=8)
        write_enable = pyrtl.Input(name="write_enable", bitwidth=1)
        mem[write_addr] <<= pyrtl.MemBlock.EnabledWrite(
            data=write_data, enable=write_enable
        )

        read_addr = pyrtl.Input(name="read_addr", bitwidth=2)
        read_data = mem[read_addr]
        read_data.name = "read_data"

        gate_graph = pyrtl.GateGraph()

        read_addr_gate = gate_graph.get_gate("read_addr")
        read_gate = gate_graph.get_gate("read_data")
        self.assertEqual(read_gate.op, "m")
        self.assertEqual(read_gate.args, [read_addr_gate])
        self.assertEqual(read_gate.op_param, (mem.id, mem))
        self.assertEqual(read_gate.bitwidth, 8)
        self.assertEqual(
            str(read_gate),
            f"read_data/8 = read(addr=read_addr/2) [memid={mem.id} mem=mem]",
        )

        write_addr_gate = gate_graph.get_gate("write_addr")
        write_data_gate = gate_graph.get_gate("write_data")
        write_enable_gate = gate_graph.get_gate("write_enable")
        write_gate = write_data_gate.dests[0]
        self.assertEqual(write_gate.op, "@")
        self.assertEqual(read_gate.op_param, (mem.id, mem))
        self.assertEqual(
            write_gate.args, [write_addr_gate, write_data_gate, write_enable_gate]
        )
        self.assertEqual(write_gate.name, None)
        self.assertEqual(write_gate.bitwidth, None)
        self.assertEqual(write_gate.dests, [])
        self.assertEqual(
            str(write_gate),
            "write(addr=write_addr/2, data=write_data/8, enable=write_enable/1) "
            f"[memid={mem.id} mem=mem]",
        )

    def test_gate_sets(self):
        a = pyrtl.Input(name="a", bitwidth=1)
        b = pyrtl.Input(name="b", bitwidth=1)

        c = pyrtl.Const(name="c", bitwidth=1, val=0)
        d = pyrtl.Const(name="d", bitwidth=1, val=1)

        x = pyrtl.Output(name="x", bitwidth=1)
        y = pyrtl.Output(name="y", bitwidth=1)

        r = pyrtl.Register(name="r", bitwidth=1)
        s = pyrtl.Register(name="s", bitwidth=1)

        mem = pyrtl.MemBlock(name="mem", bitwidth=1, addrwidth=1)

        x <<= a + c

        r.next <<= r + c
        s.next <<= r + d

        mem[a] <<= pyrtl.MemBlock.EnabledWrite(data=c, enable=d)
        read = mem[b]
        read.name = "read"
        y <<= read + d

        gate_graph = pyrtl.GateGraph()

        self.assertEqual(sorted(gate.name for gate in gate_graph.inputs), ["a", "b"])
        self.assertEqual(sorted(gate.name for gate in gate_graph.consts), ["c", "d"])
        self.assertEqual(sorted(gate.name for gate in gate_graph.outputs), ["x", "y"])
        self.assertEqual(sorted(gate.name for gate in gate_graph.registers), ["r", "s"])
        self.assertEqual(sorted(gate.name for gate in gate_graph.mem_reads), ["read"])
        mem_writes = gate_graph.mem_writes
        # Check the MemBlock write.
        self.assertEqual(len(mem_writes), 1)
        write_gate = next(iter(mem_writes))
        self.assertTrue(write_gate is not None)
        self.assertEqual(write_gate.op, "@")
        # MemBlock write has no name.
        self.assertEqual(write_gate.name, None)

        self.assertEqual(
            sorted(gate.name for gate in gate_graph.sources),
            ["a", "b", "c", "d", "r", "s"],
        )
        sinks = set(gate_graph.sinks)
        self.assertEqual(
            sorted(str(gate.name) for gate in sinks), ["None", "r", "s", "x", "y"]
        )


if __name__ == "__main__":
    unittest.main()



# Import order matters in this file.
# isort: skip_file

# error types thrown
from .pyrtlexceptions import PyrtlError, PyrtlInternalError

# core rtl constructs
from .core import (
    LogicNet,
    Block,
    PostSynthBlock,
    working_block,
    reset_working_block,
    set_working_block,
    temp_working_block,
    set_debug_mode,
)

# convenience classes for building hardware
from .wire import WireVector, Input, Output, Const, Register

from .gate_graph import GateGraph, Gate

# helper functions
from .helperfuncs import (
    input_list,
    output_list,
    register_list,
    wirevector_list,
    log2,
    truncate,
    match_bitpattern,
    bitpattern_to_val,
    chop,
    val_to_signed_integer,
    val_to_formatted_str,
    formatted_str_to_val,
    infer_val_and_bitwidth,
    probe,
    rtl_assert,
    check_rtl_assertions,
    find_loop,
    find_and_print_loop,
    wire_struct,
    wire_matrix,
    one_hot_to_binary,
    binary_to_one_hot,
)

from .corecircuits import (
    and_all_bits,
    or_all_bits,
    xor_all_bits,
    rtl_any,
    rtl_all,
    mux,
    select,
    concat,
    concat_list,
    parity,
    tree_reduce,
    as_wires,
    match_bitwidth,
    enum_mux,
    bitfield_update,
    bitfield_update_set,
    signed_add,
    signed_sub,
    signed_mult,
    signed_lt,
    signed_le,
    signed_gt,
    signed_ge,
    shift_left_arithmetic,
    shift_right_arithmetic,
    shift_left_logical,
    shift_right_logical,
)

# memory blocks
from .memory import MemBlock, RomBlock

# conditional updates
from .conditional import conditional_assignment, otherwise, currently_under_condition

# block simulation support
from .simulation import Simulation, FastSimulation, SimulationTrace, enum_name
from .compilesim import CompiledSimulation

# block visualization output formats
from .visualization import (
    output_to_trivialgraph,
    graphviz_detailed_namer,
    output_to_graphviz,
    output_to_svg,
    block_to_graphviz_string,
    block_to_svg,
    trace_to_html,
    net_graph,
)

# import from and export to file format routines
from .importexport import (
    input_from_verilog,
    output_to_verilog,
    output_verilog_testbench,
    input_from_blif,
    output_to_firrtl,
    input_from_iscas_bench,
)

# different transform passes
from .passes import (
    common_subexp_elimination,
    constant_propagation,
    synthesize,
    nand_synth,
    and_inverter_synth,
    optimize,
    one_bit_selects,
    two_way_concat,
    direct_connect_outputs,
    two_way_fanout,
)

from .transform import (
    net_transform,
    wire_transform,
    copy_block,
    clone_wire,
    replace_wires,
    replace_wire_fast,
)

# analysis and estimation functions
from .analysis import (
    area_estimation,
    TimingAnalysis,
    yosys_area_delay,
    paths,
    distance,
    fanout,
)

__all__ = [
    # pyrtlexceptions
    "PyrtlError",
    "PyrtlInternalError",
    # core
    "LogicNet",
    "Block",
    "PostSynthBlock",
    "working_block",
    "reset_working_block",
    "set_working_block",
    "temp_working_block",
    "set_debug_mode",
    # wire
    "WireVector",
    "Input",
    "Output",
    "Const",
    "Register",
    # gate_graph
    "GateGraph",
    "Gate",
    # helperfuncs
    "input_list",
    "output_list",
    "register_list",
    "wirevector_list",
    "log2",
    "truncate",
    "match_bitpattern",
    "bitpattern_to_val",
    "chop",
    "val_to_signed_integer",
    "val_to_formatted_str",
    "formatted_str_to_val",
    "infer_val_and_bitwidth",
    "probe",
    "rtl_assert",
    "check_rtl_assertions",
    "find_loop",
    "find_and_print_loop",
    "wire_struct",
    "wire_matrix",
    "one_hot_to_binary",
    "binary_to_one_hot",
    # corecircuits
    "and_all_bits",
    "or_all_bits",
    "xor_all_bits",
    "rtl_any",
    "rtl_all",
    "mux",
    "select",
    "concat",
    "concat_list",
    "parity",
    "tree_reduce",
    "as_wires",
    "match_bitwidth",
    "enum_mux",
    "bitfield_update",
    "bitfield_update_set",
    "signed_add",
    "signed_sub",
    "signed_mult",
    "signed_lt",
    "signed_le",
    "signed_gt",
    "signed_ge",
    "shift_left_arithmetic",
    "shift_right_arithmetic",
    "shift_left_logical",
    "shift_right_logical",
    # memory
    "MemBlock",
    "RomBlock",
    # conditional
    "conditional_assignment",
    "otherwise",
    "currently_under_condition",
    # simulation
    "Simulation",
    "FastSimulation",
    "SimulationTrace",
    "enum_name",
    # compilesim
    "CompiledSimulation",
    # visualization
    "output_to_trivialgraph",
    "graphviz_detailed_namer",
    "output_to_graphviz",
    "output_to_svg",
    "block_to_graphviz_string",
    "block_to_svg",
    "trace_to_html",
    "net_graph",
    # importexport
    "input_from_verilog",
    "output_to_verilog",
    "output_verilog_testbench",
    "input_from_blif",
    "output_to_firrtl",
    "input_from_iscas_bench",
    # passes
    "common_subexp_elimination",
    "constant_propagation",
    "synthesize",
    "nand_synth",
    "and_inverter_synth",
    "optimize",
    "one_bit_selects",
    "two_way_concat",
    "direct_connect_outputs",
    "two_way_fanout",
    # transform
    "net_transform",
    "wire_transform",
    "copy_block",
    "clone_wire",
    "replace_wires",
    "replace_wire_fast",
    # analysis
    "area_estimation",
    "TimingAnalysis",
    "yosys_area_delay",
    "paths",
    "distance",
    "fanout",
]
